import numpy as np
from scipy.optimize import curve_fit
from scipy.sparse.linalg import expm, eigsh, expm_multiply, svds
from scipy.sparse import csr_matrix, csc_matrix, csc_array, kron as spkron
import os
from ncon import ncon
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.patches import Ellipse
from matplotlib.animation import FuncAnimation
from typing import Union
from functools import partial

# ---------------------------------------------------------------------------------------
# Tensor shapes
# ---------------------------------------------------------------------------------------
def tensor_shapes(lists: np.ndarray, prnt: bool=True):
    """
    tensor_shapes

    This function returns the shapes of the tensors
    present in a given list.

    lists: list - list of tensors, e.g. the mps

    """
    shapes = [array.shape for array in lists]
    if prnt:
        for i in range(len(lists)):
            print(lists[i].shape)

    return shapes


# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
"""
Saving and loading tools
"""
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------
# Get precision
# ---------------------------------------------------------------------------------------
def get_precision(num: float):
    """
    get_precision

    This function finds the precision needed to save a parameter
    in a certain interval so that all the significant figures are saved.

    num: float - it is the range of the interval over the number of points in the interval

    """
    # Convert the number to a string to work with its representation
    num_str = str(num)

    # Split the number into its integer and fractional parts
    integer_part, fractional_part = num_str.split(".")

    # Count leading zeros in the fractional part
    leading_zeros = len(fractional_part) - len(fractional_part.lstrip("0"))

    # Calculate the absolute value of the exponent
    exponent = (
        leading_zeros + 1
    )  # Subtract 1 to account for the first digit before the decimal point

    return abs(exponent)


# ---------------------------------------------------------------------------------------
# Get labels
# ---------------------------------------------------------------------------------------
def get_labels(shapes):
    """
    get_labels

    This function computes the labels to split a flat list
    in arrays (still flat) corresponding to the original tensors.
    In order to do so it is sufficient to know how many elements
    where present in the original tensors. This information is given
    by the product of the shape of each tensor which is saved in shapes.
    Once we know the elements in one tensor we sum the product of
    shapes to get the correct index in the list.

    shapes: list - list of shapes of the tensors of, e.g., a mps.
    """
    labels = []
    label = 0
    for shape in shapes:
        label += np.prod(shape)
        labels.append(label)
    return labels


# ---------------------------------------------------------------------------------------
# Renaming
# ---------------------------------------------------------------------------------------
def renaming(folder, hs):
    """
    renaming

    This function can be used to rename files in a certain
    folder. If the file is initially saved by including in its name
    a float number which is not automatically approximated correctly,
    we can rename it by giving that array of floating numbers.

    folder: string - path or folder where the files have to be loaded
    hs: np.ndarray - array of floating numbers which were not saved correctly

    """
    # e.g. of folder
    # folder = "/data/fdimarca/tensor_network/bonds_data/"
    # array which is saved in the files with extra zeros
    # hs = np.arange(0.8,1.2,0.01)

    # Iterate
    for file in os.listdir(folder):
        # Taking the old name
        oldName = os.path.join(folder, file)
        # We split the filename where the point is
        n = os.path.splitext(file)[0]
        for h in hs:
            # We take the precision we are interested in saving
            h = f"{h:.2f}"
            # We split the filename where the point is
            h = os.path.splitext(h)[1]
            if h in oldName:
                b = n + h
                newName = os.path.join(folder, b)
                # print(newName)
                # Rename the file
                os.rename(oldName, newName)


# ---------------------------------------------------------------------------------------
# Save list of lists
# ---------------------------------------------------------------------------------------
def save_list_of_lists(file_path, list):
    """
    save_list_of_lists

    This function allows you to save the list of list at the file path specified

    file_path: string - file path
    list: list - list to save

    """
    with open(file_path, "w") as file:
        for sublist in list:
            line = " ".join(repr(item) for item in sublist)
            file.write(line + "\n")


# ---------------------------------------------------------------------------------------
# Load list of lists
# ---------------------------------------------------------------------------------------
def load_list_of_lists(file_path):
    """
    load_list_of_lists

    This function allows you to load the list of list from the file path specified

    file_path: string - file path

    """
    loaded_data = []

    with open(file_path, "r") as file:
        lines = file.readlines()
        for line in lines:
            # Remove square brackets and split the line into elements
            elements = line.strip("[]\n").split()
            # Convert elements to floats and remove square brackets from individual elements
            el = []
            for element in elements:
                if element.strip("[]") != "":
                    el.append(float(element.strip("[]")))
            # elements = [float(element.strip("[]")) for element in elements]

            # Append the sublist to the loaded_data list
            loaded_data.append(el)
            # loaded_data.append(elements)

    return loaded_data

# ---------------------------------------------------------------------------------------
# Access txt
# ---------------------------------------------------------------------------------------
def access_txt(file_path: str, column_index: int):
    """
    access_txt

    This function accesses to .txt files that have
    an equal number of space separated values for each row.
    We can access to one specific column of the .txt file.

    file_path: str - file path
    column_index: int - index of the column we want to retrieve

    """
    # Initialize an empty grid to store the values
    grid = []

    # Open and read the file
    with open(file_path, "r") as file:
        for line in file:
            # Split each line into a list of values, assuming values are separated by spaces
            values = line.strip().split()
            row = [float(value) for value in values]
            # Append the row to the grid
            grid.append(row)

    # Assuming 'grid' contains your data as a list of lists
    # Get the number of rows and columns in the grid
    m = len(grid)  # Number of rows
    n = len(
        grid[0]
    )  # Number of columns (assuming all rows have the same number of columns)

    # Initialize empty lists for each column
    columns = [[] for _ in range(n)]

    # Extract values column-wise
    for row in grid:
        for j in range(n):
            columns[j].append(row[j])

    # Now, 'columns' is a list of lists where columns[j] contains the values of the j-th column.
    column = columns[column_index]
    return column


# ---------------------------------------------------------------------------------------
# Replace Zeros with Nan
# ---------------------------------------------------------------------------------------
def replace_zeros_with_nan(input_list):
    # Convert the input list to a NumPy array
    arr = np.array(input_list)

    # Count the number of zeros
    num_zeros = np.count_nonzero(arr == 0)

    # Replace zeros with np.nan
    arr[arr == 0] = np.nan

    # Convert back to a Python list with np.nan values
    result_list = arr.tolist()

    return result_list, num_zeros

# ---------------------------------------------------------------------------------------
# Logarithm base d
# ---------------------------------------------------------------------------------------
def logarithm_base_d(x: float, d: float):
    """
    logarithm_base_d
    
    This function performs the change of base of the logarithm to d for the value x.

    x: float - value we want to evaulate the logarithm of
    d: float - value of the base of the logarithm

    """
    if x <= 0 or d <= 0 or d == 1:
        raise ValueError("Invalid input: x and d must be positive and d must not be equal to 1.")
    
    # Calculate the logarithms using the change of base formula
    result = np.log(x) / np.log(d)
    
    return result
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
"""
Critical exponent tools
"""
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------
# Variance
# ---------------------------------------------------------------------------------------
def variance(first_m, sm):
    """
    variance

    This function computes the variance of the distribution function
    of a local order parameter, e.g. magnetization, energy.

    first_m: float - it is the first momentum of the, e.g., magnetization
    sm: float - it is the second momentum of the, e.g., magnetization

    """
    return np.abs(sm - first_m**2)


# ---------------------------------------------------------------------------------------
# Binder's Cumulant
# ---------------------------------------------------------------------------------------
def binders_cumul(fourth_m, sm):
    """
    binders_cumul

    This function computes the binders cumulant of the distribution
    function of a local order parameter, e.g. magnetization, energy.

    fourth_m: float - it is the fourth momentum of the, e.g., magnetization
    sm: float - it is the second momentum of the, e.g., the magnetization

    """
    return 1 - fourth_m / (3 * sm**2)


# ---------------------------------------------------------------------------------------
# k values
# ---------------------------------------------------------------------------------------
def k_values(L):
    """
    k_values

    This function computes the k values present in the summation
    of the ground state formula for the PBC of the 1D transverse field Ising model.

    L: int - the number of spins present in the 1D chain

    """
    ks = []
    add = 2 * np.pi / L
    k = -(L - 1) * np.pi / L
    ks.append(k)
    for _ in range(L - 1):
        ks.append(ks[-1] + add)

    return ks


# ---------------------------------------------------------------------------------------
# Ground state
# ---------------------------------------------------------------------------------------
def ground_state(L):
    """
    ground_state

    This function computes the analytical solution of the ground state
    for the 1D transverse field Ising model with Periodic Boundary Conditions (PBC).

    L: int - the number of spins present in the 1D chain

    """
    ks = k_values(L)
    e_0 = []
    for k in ks:
        e_0.append(np.sqrt(2 + 2 * np.cos(k)))

    return -sum(e_0)


# ---------------------------------------------------------------------------------------
# Von Neumann Entropy
# ---------------------------------------------------------------------------------------
def von_neumann_entropy(s):
    """
    von_neumann_entropy

    This function computes the entanglement entropy
    given the Schmidt values of a system.

    s: np.ndarray - array of Schmidt values of a system

    """
    return -np.sum((s**2) * np.log2(s**2))


# ---------------------------------------------------------------------------------------
# Middle Schmidt Values
# ---------------------------------------------------------------------------------------
def get_middle_chain_schmidt_values(vec, where: int, bond: bool = True):
    """
    get_middle_chain_schmidt_values

    This function retrieve the schmidt values of a vector representing
    a chain of spins. The decomposition is operated in the middle of the chain.

    vec: csc_array - statevector of our system
    bond: bool - compute the middle chain schmidt values or
        the ones from all the chain (excluding the edge sites). By defalut True
    where: int - bond where we want to perform the Schmidt decomposition

    """
    L = int(np.log2(vec.shape[0]))
    sing_vals = []
    if bond:
        assert (
            1 < where < L - 1
        ), f"The decomposition can be performed only at bonds between {2} and {L-2}"

        new_shape = (2 ** (where), 2 ** (L - where))

        matrix = vec.reshape(new_shape)
        s = svds(
            matrix,
            k=(min(matrix.shape[0], matrix.shape[1]) - 2),
            return_singular_vectors=False,
        )
        # u, s, v = np.linalg.svd(matrix.toarray(), full_matrices=False)
        sing_vals.append(s)
    else:
        sub = [2] * (L - 2)
        for i in range(2, L - 1):
            new_shape = (2 ** (i), 2 ** (L - i))
            matrix = vec.reshape(new_shape)
            s = svds(
                matrix,
                k=(min(matrix.shape[0], matrix.shape[1]) - sub[i - 1]),
                return_singular_vectors=False,
                which="LM",
            )
            sing_vals.append(s)

    return sing_vals


# ---------------------------------------------------------------------------------------
# Fitting
# ---------------------------------------------------------------------------------------
def fitting(xs, results, guess):
    """
    fitting

    This function gives you back the parameters and the variance
    in fitting some given data with a function you can define in fit.
    Here we use as an example the function fitting the entanglement
    entropy which is used to extract the correlation length of the system
    at criticality, i.e. determining the phase transition.

    xs: Any - x data of the function
    results: Any - y data of the function
    guess: Any - guess values to assign for the optimization

    """
    assert len(xs) == len(
        results
    ), f"The x and y must have the same dimension, but x has dim({len(xs)}) and y has dim({len(results)})"

    # define the function to fit

    # fit for correlation length
    def fit(x, c, corr_length):
        return c / 6 * np.log(x - np.log(x / corr_length + np.exp(-x / corr_length)))

    # fit for computational time (chi)
    def fit(x, a, b, theta):
        return a * np.exp(theta * x) + b

    # fit your data with a given guess
    param_opt, covar_opt = curve_fit(fit, xs, results, guess)
    return param_opt, covar_opt


def mps_to_vector(mps):
    D = mps[0].shape[0]
    d = mps[0].shape[1]
    a = np.zeros(D)
    a[0] = 1
    a = ncon([a, mps[0]], [[1], [1, -1, -2]])
    for i in range(1, len(mps)):
        a_index = list(range(-i, 0))
        a_index.reverse()
        a_index.append(1)
        a = ncon([a, mps[i]], [a_index, [1, -(i + 1), -(i + 2)]])

    a_index = list(range(-i - 1, 0))
    a_index.reverse()
    a_index.append(1)
    b = np.zeros(D)
    b[-1] = 1
    final_vec = ncon([a, b.T], [a_index, [1]])
    final_vec = final_vec.reshape(d ** len(mps))
    return final_vec


def mpo_to_matrix(mpo):
    v_l = np.zeros(mpo[0].shape[0])
    v_l[0] = 1
    L = len(mpo)
    env = v_l

    mid = [1]
    label_env = mid
    for i in range(L):
        label_mpo = [1, -L * 100, -(i + 1), -(L + i + 1)]
        env = ncon([env, mpo[i]], [label_env, label_mpo])
        up = [int(-elem) for elem in np.linspace(1, i + 1, i + 1)]
        down = [int(-elem) for elem in np.linspace(L + 1, L + 1 + i, i + 1)]
        label_env = up + down + mid

    v_r = np.zeros(mpo[0].shape[0])
    v_r[-1] = 1
    d = mpo[0].shape[2]
    matrix = ncon([env, v_r.T], [label_env, mid]).reshape((d**L, d**L))
    return matrix

def swap_columns(matrix, tensor: bool=True):
    n = len(matrix)
    for i in range(n // 2):
        col_to_swap = n - 1 - i
        if tensor:
            matrix[i, :, :], matrix[col_to_swap, :, :] = matrix[col_to_swap, :, :], matrix[i, :, :].copy()
        else:
            for row in matrix:
                row[i], row[col_to_swap] = row[col_to_swap], row[i]
    return matrix

def swap_rows(matrix, tensor: bool=True):
    n = len(matrix)
    for i in range(n // 2):
        row_to_swap = n - 1 - i
        if tensor:
            matrix[:, i, :], matrix[:, row_to_swap, :] = matrix[:, row_to_swap, :], matrix[:, i, :].copy()
        else:
            for row in matrix:
                row[i], row[row_to_swap] = row[row_to_swap], row[i]
    return matrix   

# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
"""
Defining a series of functions to have single and double site operators
"""
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------
# Before
# ---------------------------------------------------------------------------------------
def before(next: csr_matrix):
    """
    before

    This function computes the tensor product between identity
    and whatever other operator specified by next.

    next: np.ndarray - Usually choose the operator that you want to add after
            an identity, e.g. I,next = I,X,I,I

    """
    id = np.eye(2)
    id = csr_matrix(id)
    return spkron(id, next)


# ---------------------------------------------------------------------------------------
# Before Tot
# ---------------------------------------------------------------------------------------
def before_tot(next: csr_matrix, site: int):
    """
    before_tot

    This function computes the tensor product between the subspace
    before the operator at a give site and the what comes after.

    next: np.ndarray - Usually choose the operator that you want to add after
            an identity, e.g. I,next = I,X,I,I
    site: int - The site where the operator is acting. Indicates the number of
            identities to be applied before

    """
    for i in range(site - 1):
        next = before(next)
    return next


# ---------------------------------------------------------------------------------------
# After
# ---------------------------------------------------------------------------------------
def after(op: csr_matrix, site: int, L: int):
    """
    after

    This function takes into account on which site the operator is acting
    and how many site there are left after it to leave unaltered,
    e.g. we act with identity.

    op: np.ndarray - operator defining our single site generalized operator
    site: int - The site where the operator is acting
    L: int - Chain length. The difference (L-site) indicates the number of
            identities to be applied after the operator op
    """
    next = np.eye(2)
    next = csr_matrix(next)
    if site < L:
        for i in range(site, L - 1):
            next = before(next)
        return spkron(op, next)
    else:
        return op


# ---------------------------------------------------------------------------------------
# Single site operator
# ---------------------------------------------------------------------------------------
def single_site_op(op: Union[csr_matrix, np.ndarray], site: int, L: int):
    """
    single_site_op

    This function combines the identities before the operator with the
    op+identities after. E.g., with after we compute op,I,···,I
    and we add to the identities before I,···,I,op,I,···,I.

    op: np.ndarray - operator defining our single site generalized operator
    site: int - The site where the operator is acting
    L: int - Chain length. The difference (L-site) indicates the number of
            identities to be applied after the operator op

    """
    assert isinstance(
        op, (csr_matrix, np.ndarray)
    ), "Input must be a NumPy array or a sparse CSR op"

    if isinstance(op, np.ndarray):
        # Convert the NumPy array to a CSR matrix
        op = csr_matrix(op)
    op_tot = before_tot(after(op, site, L), site=site)
    return op_tot


# ---------------------------------------------------------------------------------------
# Double After
# ---------------------------------------------------------------------------------------
def double_after(op: Union[csr_matrix, np.ndarray], site: int, L: int):
    """
    double_after

    This function takes into account on which site the two-site operator
    is acting and how many site there are left after it to leave unaltered,
    e.g. we act with identity.

    op: np.ndarray - operator defining our two-site generalized operator
    site: int - The tuple (site,site+1) is where the operator is acting
    L: int - Chain length. The difference (L-site+1) indicates the number of
            identities to be applied after the operator op

    """
    assert isinstance(
        op, (csr_matrix, np.ndarray)
    ), "Input must be a NumPy array or a sparse CSR matrix"

    if isinstance(op, np.ndarray):
        # Convert the NumPy array to a CSR matrix
        op = csr_matrix(op)

    next = np.eye(2)
    next = csr_matrix(next)
    if site <= L - 2:
        for i in range(site + 1, L - 1):
            next = before(next)
        next = spkron(op, next)
        return spkron(op, next)
    else:
        return spkron(op, op)


# ---------------------------------------------------------------------------------------
# Two-site operator
# ---------------------------------------------------------------------------------------
def two_site_op(op: csr_matrix, site: int, L: int):
    """
    two_site_op

    This function combines the identities before the operator with the
    op+identities after. E.g., with after we compute op,op,I,···,I
    and we add to the identities before I,···,I,op,op,I,···,I.

    op: np.ndarray - operator defining our two-site generalized operator
    site: int - The tuple (site,site+1) is where the operator is acting
    L: int - Chain length. The difference (L-site+1) indicates the number of
            identities to be applied after the operator op

    """
    assert site < L, "Site out of bounds"

    op_tot = before_tot(double_after(op, site, L), site)
    return op_tot


# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
"""
With a single and double site operators we can define different Hamiltonians
and other special operators, e.g. flipping half of the spins' chain
"""
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------
# Interaction Hamiltonian
# ---------------------------------------------------------------------------------------
def H_int(L: int, op: csr_matrix):
    """
    H_int

    This function sums the two-site interactions terms of a given operator.

    L: int - Total number of spins in the chain
    op: np.ndarray - operator defining the two-site interaction

    """
    h = 0
    for site in range(1, L):
        h += two_site_op(op, site, L)
    return h


# ---------------------------------------------------------------------------------------
# Local Hamiltonian
# ---------------------------------------------------------------------------------------
def H_loc(L: int, op: csr_matrix):
    """
    H_int

    This function sums the single-site local terms of a given operator.

    L: int - Total number of spins in the chain
    op: np.ndarray - operator defining the single-site local term

    """
    h = 0
    for site in range(1, L + 1):
        h += single_site_op(op, site, L)
    return h


# ---------------------------------------------------------------------------------------
# Ising Hamiltonian
# ---------------------------------------------------------------------------------------
def H_ising_gen(
    L: int, op_l: csr_matrix, op_t: csr_matrix, J: float, h_l: float, h_t: float
) -> csr_matrix:
    """
    H_ising_gen

    This function finds the ising hamiltonian by summing local and
    interaction terms with given parameters and operators.

    L: int - Total number of spins in the chain
    op_l: np.ndarray - operator defining the two-site interaction term
    op_t: np.ndarray - operator defining the single-site local term
    J: float - couppling constant of the interaction term
    h_l: float - couppling constant of the local logitudinal term
    h_t: float - couppling constant of the local transverse term
    """
    return -J * H_int(L, op=op_l) - h_l * H_loc(L, op_l) - h_t * H_loc(L, op_t)


# ---------------------------------------------------------------------------------------
# Flipping half
# ---------------------------------------------------------------------------------------
def flipping_half(L: int, op: csr_matrix):
    """
    flipping_half

    This function flips the half of the spin states to the right of the chain.
    To achieve the flipping it is important a correct choice of the operator

    L: int - chain size
    op: np.ndarray - flipping operator. E.g. spin in Z basis -> op=X

    """
    O = op
    for _ in range(L // 2 - 1):
        O = spkron(op, O)

    for _ in range(L // 2):
        O = before(O)
    return O


# ---------------------------------------------------------------------------------------
# Truncation
# ---------------------------------------------------------------------------------------
def truncation(array: Union[np.ndarray, csc_matrix, csr_matrix], threshold: float):
    """
    truncation

    This function truncates the entries of an array according to the preselected
    threshold.

    array: np.ndarray - array to truncate
    threshold: float - level of the truncation

    """
    if isinstance(array, np.ndarray):
        return np.where(np.abs(np.real(array)) > threshold, array, 0)
    if isinstance(array, csc_matrix) or isinstance(array, csr_matrix):
        # Apply the thresholding operation on the csc_matrix
        filtered_matrix = csc_matrix((array.shape), dtype=array.dtype)
        # Apply the thresholding operation
        filtered_matrix.data = array.data * (array.data > threshold)
        return filtered_matrix


# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
"""
Visualization tools
"""
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------
# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------
# Plot Side By Side
# ---------------------------------------------------------------------------------------
def plot_side_by_side(
    data1, data2, cmap="viridis", title1="Imshow 1", title2="Imshow 2"
):
    """
    plot_side_by_side

    This visualization function creates two plots of colormaps one next to the other.

    data1: data of the left colormap
    data2: data of the right colormap
    cmap: string - colormap chosen to visualize the two colormap. Should be the same. By default 'viridis'
    title1: string - title of the left colormap
    title2: string - title of the right colormap

    """
    # Create a figure and subplots
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 5))

    # Plot the imshow plots on the subplots
    im1 = ax1.imshow(data1, cmap=cmap, vmin=-1, vmax=1, aspect="auto")
    im2 = ax2.imshow(data2, cmap=cmap, vmin=-1, vmax=1, aspect="auto")

    # Set titles for the subplots
    ax1.set_title(title1)
    ax2.set_title(title2)

    # Remove ticks from the colorbar subplot
    # ax2.set_xticks([])
    ax2.set_yticks([])

    # Create a colorbar for the second plot on the right
    cbar = fig.colorbar(im2, ax=ax2)
    cbar = fig.colorbar(im1, ax=ax1)

    # Adjust layout and display the plot
    plt.tight_layout()
    plt.show()


# ---------------------------------------------------------------------------------------
# Create Sequential Colors
# ---------------------------------------------------------------------------------------
def create_sequential_colors(num_colors, colormap_name: str='viridis'):
    """
    create_sequential_colors

    This function creates a sequence of colors extracted from a specified colormap.

    num_colors: int - number of colors we want to extract
    colormap_name: string - colormap name we want to use

    """
    colormap = plt.cm.get_cmap(colormap_name)
    colormap_values = np.linspace(0, 1, num_colors)
    colors = [colormap(value) for value in colormap_values]
    return colors


def plot_results_DMRG(
    title: str,
    for_array: list,
    interval: int,
    fname: str,
    path: str,
    fname_ex: str,
    path_ex: str,
    fname_save: str,
    path_save: str,
    ylabel: str,
    yscale: str = "linear",
    xscale: str = "linear",
    delta: float = None,
    time: bool = False,
    exact: bool = False,
    save: bool = True,
    marker: str = "+",
    m_size: int = 25,
    linewidth: float = 1,
    alpha: float = 1,
    n_points: float = 1,
    cmap: str = "viridis",
    txt: bool = False,
    precision: int = 2,
    show: bool = True
):
    """
    plot_results_evolution

    This funciton plots the results of a time evolution for a specific model.

    """
    colors = create_sequential_colors(num_colors=len(for_array), colormap_name=cmap)

    plt.title(
        title,
        fontsize=14,
    )
    step = int(1 // n_points)
    x = interval[::step]

    for i, elem in enumerate(for_array):
        if txt:
            res_mps = np.loadtxt(f"{path}/{fname}_chi_{elem}")
        else:
            res_mps = np.load(f"{path}/{fname}_chi_{elem}.npy")
        # res_mps = np.loadtxt(f"{path}/{fname}_chi_{elem}")
        # res_mps = access_txt(
        #     f"{path}/all_bond_entropy_Ising_L_51_flip_True_delta_0.01_chi_{elem}_h_ev_1.75", 25
        # )
        y = res_mps[::step]
        # first_elem = np.array(0)
        # y = np.append(first_elem, y)[::step]
        plt.scatter(
            x,
            y,
            s=m_size,
            marker=marker,
            alpha=alpha,
            linewidths=linewidth,
            facecolors=colors[i],
            label=f"mps: $\\chi={elem}$",
        )

        
    # labels = interval[:: (len(interval) // 10)],
    # labels = [f"{h:.{precision}f}" for h in labels]
    

    if time:
        plt.xlabel(f"time $(\\delta t)$")
        labels = delta * np.asarray(interval[:: (len(interval) // 5)])
        labels = [round(lab, 1) for lab in labels]
        plt.xticks(
            ticks=np.asarray(interval[:: (len(interval) // 5)]),
            labels=labels,
        )
    else:
        plt.xlabel("plaquette term (h)")
        labels = np.asarray(interval[:: (len(interval) // 5)])
        labels = [round(lab, 1) for lab in labels]
        plt.xticks(
            ticks=interval[:: (len(interval) // 5)],
            labels=labels,
        )
    plt.ylabel(ylabel)
    if exact:
            # res_exact = np.loadtxt(f"{path_ex}/{fname_ex}")
            res_exact = np.load(f"{path_ex}/{fname_ex}.npy")
            res_exact = res_exact[::step]
            plt.plot(
                x,
                res_exact,
                color="indianred",
                linewidth=0.6,
                label=f"exact",
            )
    plt.legend()
    plt.yscale(yscale)
    plt.xscale(xscale)
    if save:
        plt.savefig(f"{path_save}/{fname_save}_marker_{marker}.png")
    if show:
        plt.show()
    plt.close()


def plot_results_TEBD(
    title: str,
    for_array: list,
    trotter_steps: int,
    delta: float,
    second_part: str,
    fname: str,
    path: str,
    fname_ex: str,
    path_ex: str,
    fname_save: str,
    path_save: str,
    ylabel: str,
    exact: bool = False,
    save: bool = True,
    marker: str = "+",
    m_size: int = 25,
    linewidth: float = 1,
    alpha: float = 1,
    n_points: float = 1,
    cmap: str = "viridis",
):
    """
    plot_results_evolution

    This funciton plots the results of a time evolution for a specific model.

    """
    colors = create_sequential_colors(num_colors=len(for_array), colormap_name=cmap)

    plt.title(
        title,
        fontsize=14,
    )
    step = 1 // n_points
    x = list(np.arange(trotter_steps + 1))[::step]

    for i, elem in enumerate(for_array):
        res_mps = np.loadtxt(f"{path}/{fname}_chi_{elem}{second_part}")
        y = res_mps[::step]
        plt.scatter(
            x,
            y,
            s=m_size,
            marker=marker,
            alpha=alpha,
            linewidths=linewidth,
            facecolors=colors[i],
            label=f"mps: $\\chi={elem}$",
        )

        if exact:
            res_exact = np.loadtxt(f"{path_ex}/{fname_ex}_chi_{elem}")
            res_exact = res_exact[::step]
            plt.plot(
                x,
                res_exact,
                color="indianred",
                label=f"exact",
            )
        plt.xlabel("time (t)")
        plt.ylabel(ylabel)
        plt.xticks(
            ticks=x[:: int(len(x) / 5)],
            labels=list(delta * np.asarray(x))[:: int(len(x) / 5)],
        )
        plt.legend()

    if save:
        plt.savefig(f"{path_save}/{fname_save}.png")
    plt.show()


def plot_colormaps_evolution(
    title: str,
    fname: str,
    path: str,
    fname_save: str,
    path_save: str,
    xlabel: str,
    xticks: np.ndarray,
    xlabels: np.ndarray,
    yticks: np.ndarray,
    ylabels: np.ndarray,
    X,
    Y,
    save: bool = True,
    cmap: str = "viridis",
    interpolation: str = "antialiased",
    d: bool = False,
    view_init: bool = False,
):
    matrix = np.loadtxt(f"{path}/{fname}")
    print(matrix.shape)
    print(X.shape)
    print(Y.shape)
    if d:
        X = X[:-1, :]
        Y = Y[:-1, :]
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.set_title(title, fontsize=14)
        ax.plot_surface(X, Y, matrix, cmap=cmap)
        ax.set_xticks(ticks=xticks, labels=xlabels)
        ax.set_xlabel(xlabel)
        ax.set_yticks(ticks=yticks, labels=ylabels)
        ax.set_ylabel("time (t)")
        if view_init:
            ax.view_init(20, 80)
        if save:
            fig.savefig(f"{path_save}/{fname_save}_3D.png")
        fig.show()
    else:
        plt.title(title, fontsize=14)
        plt.imshow(matrix, cmap=cmap, aspect="auto", interpolation=interpolation)
        plt.colorbar()
        plt.xticks(ticks=xticks, labels=xlabels)
        plt.xlabel(xlabel)
        plt.yticks(ticks=yticks, labels=ylabels)
        plt.ylabel("time (t)")

    if save:
        plt.savefig(f"{path_save}/{fname_save}.png")
    plt.show()


def anim(frames: int, interval: int, data: np.ndarray, params: np.ndarray, show: bool, charges_x: list, charges_y: list, precision: int, time: bool):

    # Create a figure and axis
    fig, ax = plt.subplots()
    title = ax.set_title("")

    # create the lattice
    hlines = list(range(data[0].shape[0]))[::2]
    vlines = list(range(data[0].shape[1]))[::2] 
    ax.hlines(y=hlines, xmin=0, xmax=data[0].shape[1]-1, colors='k', linewidth=0.8)
    ax.vlines(x=vlines, ymin=0, ymax=data[0].shape[0]-1, colors='k', linewidth=0.8)

    # add the links expectation values
    cmap = plt.get_cmap('viridis')
    im = ax.imshow(data[0], vmin=-1, vmax=1, cmap=cmap, interpolation="nearest")
    cbar = plt.colorbar(im, ax=ax)
    cbar.set_label("Electric field")

    # add vacuum
    for i in vlines:
        for j in hlines:
            circle = Ellipse((i, j), 0.1, 0.1, edgecolor='blue', facecolor='none', linewidth=1)
            ax.add_patch(circle)

    # add charges
    if charges_x != None:
        for i in range(len(charges_x)):
            ax.text(x=2*charges_x[i]+0.1, y=2*charges_y[i]-0.1, s="-1", color="red")
            circle = Ellipse((2*charges_x[i], 2*charges_y[i]), 0.1, 0.1, edgecolor='red', facecolor='none', linewidth=1)
            ax.add_patch(circle)

    
    # Function to update the colormap in each frame
    def update(frame, data: np.ndarray, params: np.ndarray, precision: int, time: bool):
        # print(frame, type(frame))
        # Generate some example data
        data_frame = data[frame]
        param_frame = params[frame]

        # Update the colormap
        im.set_data(data_frame)
        # im.imshow(data_frame, vmin=0, vmax=1, cmap=cmap, interpolation="nearest")
        if time:
            title.set_text(f'Trotter step: {param_frame:.{precision}f}')
        else:
            title.set_text(f'Magnetic term: {param_frame:.{precision}f}')
        # Set colorbar
        # cbar.set(im, ax=ax)
        

    # Create the animation
    animation = FuncAnimation(fig, partial(update, data=data, params=params, precision=precision, time=time), frames=frames, interval=interval, repeat=False)

    # Show the animation
    if show:
        plt.show()
    return animation



from scipy.sparse import csc_array, identity, linalg
from ncon import ncon
from qs_mps.sparse_hamiltonians_and_operators import sparse_pauli_z, sparse_pauli_x
from qs_mps.utils import mpo_to_matrix, tensor_shapes
from qs_mps.lattice import Lattice
from qs_mps.mps_class import MPS
import numpy as np


class MPO_ladder:
    def __init__(self, L, l, model=str, lamb=None):
        self.L = L
        self.l = l
        self.model = model
        self.charges = np.ones((l + 1, L))
        self.lamb = lamb
        self.mpo = []
        self.latt = self.latt = Lattice((self.L, self.l + 1), (False, False))
        self.dof = self.l * (self.L - 1)

    def charge_constraint(self):
        """
        charge_constraint

        We decide arbitrarily to impose a charge constraint on the bottom right
        corner of the direct lattice.

        """
        # take the product of the first l-1 rows
        productory = np.prod([self.charges[line, :] for line in range(self.l)])
        # take the last row but the last charge
        productory = productory * np.prod([self.charges[-1, :-1]])
        self.charges[-1, -1] = productory
        return self

    def add_charges(self, rows: list, columns: list):
        """
        add_charges

        This function adds the charges to the background
        vacuum sector (all positive charges). The number of
        charges we add are given by the len of each indices list

        rows: list - row indices of the charges to add
        columns: list - column indices of the charges to add
        """
        for i, j in zip(rows, columns):
            assert ((i != -1) and (j != -1)) or (
                (i == self.l - 1) and (j == self.L - 1)
            ), "Do not choose the last charge! We use it for Gauss Law constraint"
            self.charges[j, i] = -1
        # impose the constraint
        self.charge_constraint()
        self.charges = np.flip(self.charges, axis=1)
        return self

    def mpo_skeleton(self):
        """
        mpo_skeleton

        This function initializes the mpo tensor or shape (2+l,2+l,2**l,2**l)
        with O matrices. We add as well the identities in the first and last
        element of the mpo tensor.

        """
        I = identity(2**self.l, dtype=complex)
        O = csc_array((2**self.l, 2**self.l), dtype=complex)
        skeleton = np.array(
            [[O.toarray() for i in range(2 + self.l)] for j in range(2 + self.l)]
        )
        skeleton[0, 0] = I.toarray()
        skeleton[-1, -1] = I.toarray()
        self.mpo = skeleton
        return self

    def charge_coeff_interaction(self, n: int, mpo_site: int):
        """
        charge_coeff_interaction

        This function computes the coefficient for the interaction in the
        ladder Z2 hamiltonian.

        The coefficient

                c_n(j,j+1) = \prod_{k=(n+1)}^{l+1} q_{k,(n+1)}

        is the product of the charges in the direct lattice "below"
        the interacting sites of the dual lattice.

        n: int - goes from 1 to l and indicate in which row of the dual lattice
            is happening the interaction
        mpo_site: int - indicates for which mpo we are computing the coefficient.
            There are (self.L-1) mpos and everyone encodes the interaction term with a
            specific charge coefficient. We take the charges starting from
            the row below the one of the interaction. Accepted values are from
            0 to (self.L - 2)

        """
        assert 1 <= n <= self.l, "Select a value of n within (1,l)"
        assert (
            0 <= mpo_site <= (self.L - 2)
        ), "Select a value of charge column within (0,L-2)"

        if mpo_site == (self.L - 2):
            c_n_j = 1
        else:
            c_n_j = np.prod(self.charges[n:, mpo_site + 1])
        return c_n_j

    def charge_coeff_local_Z(self, n: int, mpo_site: int):
        """
        charge_coeff_local_Z

        This function computes the coefficient for the local Z term in the
        ladder Z2 hamiltonian.

        """
        assert 1 <= n <= self.l, "Select a value of n within (1,l)"

        if mpo_site == 0 or mpo_site == (self.L - 2):
            if mpo_site == 0:
                col = mpo_site
                if n == 1:
                    alpha = 1
                    n = n - 1
                elif n == self.l:
                    alpha = 1
                    n = n
                else:
                    alpha = 0
            elif mpo_site == (self.L - 2):
                col = mpo_site + 1
                if n == 1:
                    alpha = 1
                    n = n - 1
                elif n == self.l:
                    alpha = 1
                    n = n
                else:
                    alpha = 0

            c_n_j = (1 + self.charges[n, col]) ** (alpha) * np.prod(
                self.charges[:n, col]
            )
        else:
            if n == 1:
                c_n_j = np.prod([self.charges[:, col] for col in range(mpo_site + 1)])
            elif n == self.l:
                c_n_j = 1
            else:
                c_n_j = 0

        return c_n_j

    def mpo_Z2_ladder_generalized(self):
        """
        mpo_Z2_ladder_generalized

        This function computes the hamiltonian MPO for the Z2
        gauge pure theory for general number of ladders and charge configuration.

        """
        self.mpo_skeleton()
        mpo_list = []
        for mpo_site in range(self.L - 1):
            # -----------
            # first row
            # -----------
            # interaction terms (from column 2 to l+1)
            for n in range(1, self.l + 1):
                # self.mpo[0,n] = - 1 * sparse_pauli_z(n=n-1, L=self.l).toarray()
                self.mpo[0, n] = (
                    -self.charge_coeff_interaction(n=n, mpo_site=mpo_site)
                    * sparse_pauli_z(n=n - 1, L=self.l).toarray()
                )

            # local terms (column l+2)
            # local Z and local X
            for i in range(1, self.l + 1):
                # self.mpo[0,n+1] += - 3 * sparse_pauli_z(n=i-1, L=self.l).toarray() - self.lamb * sparse_pauli_x(n=i-1, L=self.l).toarray()
                self.mpo[0, n + 1] += (
                    -self.charge_coeff_local_Z(n=i, mpo_site=mpo_site)
                    * sparse_pauli_z(n=i - 1, L=self.l).toarray()
                    - self.lamb * sparse_pauli_x(n=i - 1, L=self.l).toarray()
                )
            # vertical Z interaction
            for j in range(self.l - 1):
                self.mpo[0, n + 1] += -(
                    sparse_pauli_z(n=j, L=self.l) @ sparse_pauli_z(n=j + 1, L=self.l)
                ).toarray()

            # -----------
            # last column
            # -----------
            # positive terms to complete the interaction between nearest neighbor sites (from row 2 to l+1)
            for n in range(1, self.l + 1):
                self.mpo[n, -1] = sparse_pauli_z(n=n - 1, L=self.l).toarray()

            mpo_list.append(self.mpo)
            self.mpo_skeleton()

        self.mpo = mpo_list
        return mpo_list

    def diagonalize(self):
        self.mpo_Z2_ladder_generalized()
        H = mpo_to_matrix(self.mpo)
        e, v = np.linalg.eigh(H)
        return e, v

    def mpo_Z2_quench_global(self, l: int = None, delta: float = None, mps: MPS = None):
        self._initialize_quench_local()
        mps.mpo_to_mps()
        self.mpo_Z2_quench_ladder(l=l, delta=delta)

    def _initialize_quench_local(self):
        I = identity(2**self.l, dtype=complex).toarray()
        w_tot = []
        for _ in range(self.L - 1):
            w_init_X = I
            for l in range(self.l):
                X_l = sparse_pauli_x(n=l, L=self.l).toarray()
                w_init_X_l = linalg.expm(1j * self.lamb * X_l)
                w_init_X = ncon([w_init_X_l, w_init_X], [[-1, 1], [1, -2]])
            w_tot.append(w_init_X)
        self.mpo = w_tot
        return self

    def mpo_Z2_quench_ladder(self, l, delta):
        I = identity(2**self.l, dtype=complex).toarray()
        Z = sparse_pauli_z(n=l - 1, L=self.l).toarray()
        Z_ll = (
            sparse_pauli_z(n=l - 1, L=self.l) @ sparse_pauli_z(n=l, L=self.l)
        ).toarray()
        w_int_loc = np.array(linalg.expm(1j * delta / 4 * Z_ll))

        if l != 1 and l != self.l:
            Z_ll_prime = (
                sparse_pauli_z(n=l, L=self.l) @ sparse_pauli_z(n=l + 1, L=self.l)
            ).toarray()
            w_int_loc_prime = np.array(linalg.expm(1j * delta / 4 * Z_ll_prime))

        w_tot = []
        for mpo_site in range(self.L - 1):
            c_loc = self.charge_coeff_interaction(n=l - 1, mpo_site=mpo_site)
            w_loc = np.array(linalg.expm(1j * c_loc * delta / 2 * Z))

            c_int = self.charge_coeff_interaction(n=l, mpo_site=mpo_site)
            w_even = np.array(
                [
                    [
                        np.sqrt(np.cos(c_int * delta)) * I,
                        1j * np.sqrt(np.sin(c_int * delta)) * Z,
                    ]
                ]
            )
            w_odd = np.array(
                [
                    [
                        np.sqrt(np.cos(c_int * delta)) * I,
                        np.sqrt(np.sin(c_int * delta)) * Z,
                    ]
                ]
            )
            w_odd = np.swapaxes(w_odd, axis1=0, axis2=1)
            if mpo_site == 0:
                if l != 1 and l != self.l:
                    w_in = ncon(
                        [
                            w_loc,
                            w_int_loc,
                            w_int_loc_prime,
                            w_even,
                            w_int_loc_prime,
                            w_int_loc,
                            w_loc,
                        ],
                        [
                            [-3, 5],
                            [5, 3],
                            [3, 1],
                            [-1, -2, 1, 2],
                            [2, 4],
                            [4, 6],
                            [6, -4],
                        ],
                    )
                else:
                    w_in = ncon(
                        [w_loc, w_int_loc, w_even, w_int_loc, w_loc],
                        [[-3, 3], [3, 1], [-1, -2, 1, 2], [2, 4], [4, -4]],
                    )
                w_tot.append(w_in)

            elif mpo_site == self.L - 1:
                if l != 1 and l != self.l:
                    w_fin = ncon(
                        [
                            w_loc,
                            w_int_loc,
                            w_int_loc_prime,
                            w_odd,
                            w_int_loc_prime,
                            w_int_loc,
                            w_loc,
                        ],
                        [
                            [-3, 5],
                            [5, 3],
                            [3, 1],
                            [-1, -2, 1, 2],
                            [2, 4],
                            [4, 6],
                            [6, -4],
                        ],
                    )
                else:
                    w_fin = ncon(
                        [w_loc, w_int_loc, w_odd, w_int_loc, w_loc],
                        [[-3, 3], [3, 1], [-1, -2, 1, 2], [2, 4], [4, -4]],
                    )

            else:
                if l != 1 and l != self.l:
                    w = ncon(
                        [
                            w_loc,
                            w_int_loc,
                            w_int_loc_prime,
                            w_even,
                            w_odd,
                            w_int_loc_prime,
                            w_int_loc,
                            w_loc,
                        ],
                        [
                            [-5, 8],
                            [8, 6],
                            [6, 4],
                            [-1, -3, 4, 3],
                            [-2, -4, 3, 5],
                            [5, 7],
                            [7, 9],
                            [9, -6],
                        ],
                    ).reshape(
                        w_odd.shape[0] * w_even.shape[0],
                        w_odd.shape[1] * w_even.shape[1],
                        w_odd.shape[2],
                        w_even.shape[3],
                    )
                else:
                    w = ncon(
                        [w_loc, w_int_loc, w_even, w_odd, w_int_loc, w_loc],
                        [
                            [-5, 6],
                            [6, 4],
                            [-1, -3, 4, 3],
                            [-2, -4, 3, 5],
                            [5, 7],
                            [7, -6],
                        ],
                    ).reshape(
                        w_odd.shape[0] * w_even.shape[0],
                        w_odd.shape[1] * w_even.shape[1],
                        w_odd.shape[2],
                        w_even.shape[3],
                    )
            w_tot.append(w)

        w_tot.append(w_fin)
        self.mpo = w_tot
        return self


# mpo = MPO_ladder(l=2, L=3, lamb=0)
# rows = []
# columns = []
# mpo.add_charges([1,1],[0,2])
# print(f"charges:\n{mpo.charges}")
# # print(f"degrees of freedom:\n{mpo.dof}")
# print(f"lattice:\n{mpo.latt._lattice_drawer.draw_lattice()}")
# # print(f"shape:\n{tensor_shapes(mpo.mpo)}")
# e, v = mpo.diagonalize()
# print(f"spectrum:\n{e}")



from qs_mps.utils import *
import matplotlib.pyplot as plt
import argparse

# default parameters of the plot layout
plt.rcParams["text.usetex"] = True  # use latex
plt.rcParams["font.size"] = 13
plt.rcParams["figure.dpi"] = 300
plt.rcParams["figure.constrained_layout.use"] = True

parser = argparse.ArgumentParser(prog="plot animations")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values or number of trotter steps",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("o", help="Observable we want to compute. Available are 'wl', 'el'", type=str)
parser.add_argument("-L", "--Ls", help="Number of rungs per ladder", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument(
    "-f", "--h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "-d", "--delta", help="Width of each time slice during the time evolution. Should be 'small enough'", type=float
)
parser.add_argument(
    "-ev", "--h_ev", help="Quench value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument("-lx","--sites", help="Number of sites in the wilson loop", type=int)
parser.add_argument("-ly","--ladders", help="Number of ladders in the wilson loop", type=int)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
)
parser.add_argument(
    "-U", "--gauss", help="Gauss constraint parameter", default=1e+3, type=float
)
parser.add_argument(
    "-sh",
    "--show",
    help="Show the animation. By default False",
    action="store_true",
)
parser.add_argument(
    "-v",
    "--save",
    help="Save the animation. By default True",
    action="store_false",
)
parser.add_argument(
    "-e",
    "--exact",
    help="Exact computation. By default False",
    action="store_true",
)
parser.add_argument(
    "-t",
    "--time",
    help="Animation of real time evolution of the observable. By default False",
    action="store_true",
)
parser.add_argument(
    "-i",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)

args = parser.parse_args()

# define the interval of equally spaced values of external field or trotter steps
if args.time:
    interval = range(args.npoints)
    precision = get_precision(args.h_i)
else:
    if args.interval == "lin":
        interval = np.linspace(args.h_i, args.h_f, args.npoints)
    elif args.interval == "log":
        interval = np.logspace(args.h_i, args.h_f, args.npoints)

    # define the precision
    num = (args.h_f - args.h_i) / args.npoints
    precision = get_precision(num)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_state = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_state = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_state = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")




for L in args.Ls:
    # define the sector by looking of the given charges
    if len(args.charges_x) == 0:
        sector = "vacuum_sector"
        charges_x = None
        charges_y = None
    else:
        sector = f"{len(args.charges_x)}_particle(s)_sector"
        charges_x = args.charges_x
        charges_y = args.charges_y

    for chi in args.chis:
        if args.o == 'el':
            
            if args.exact:
                path_file = f"electric_field_{args.model}_direct_lattice_{args.l-1}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}"
                dataname = f"{parent_path}/results/exact/electric_field/{path_file}.npy"
                savename = f"{parent_path}/figures/exact/animations/animation_{path_file}.mp4"
            elif args.time:
                path_file = f"electric_field_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}"
                dataname = f"{parent_path}/results/electric_field/{path_file}.npy"
                savename = f"{parent_path}/figures/animations/animation_{path_file}.mp4"
            else:
                path_file = f"electric_field_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}"
                dataname = f"{parent_path}/results/electric_field/{path_file}.npy"
                savename = f"{parent_path}/figures/animations/animation_{path_file}.mp4"
            
            data = np.load(dataname)


        movie = anim(frames=args.npoints, interval=200, data=data, params=interval, show=args.show, charges_x=args.charges_x, charges_y=args.charges_y, precision=precision, time=args.time)
        if args.save:
            movie.save(savename)


from qs_mps.mps_class import MPS
from qs_mps.utils import tensor_shapes
import numpy as np

L = 12
chain = MPS(L=L, d=2, model="ANNNI", chi=64, h=0, k=0, J=1, eps=1e-5)

up_state = [1]
down_state = [0]
ferro_state = np.array([[up_state,down_state]])
ferro_tensor = [ferro_state for _ in range(L)]
chain.sites = ferro_tensor.copy()
chain.enlarge_chi()
tensor_shapes(chain.sites)
# chain.canonical_form()
# tensor_shapes(chain.sites)
rdm = chain.reduced_density_matrix(sites=[L//2])


# import packages
import argparse
from qs_mps.mps_class import MPS
from qs_mps.utils import *

parser = argparse.ArgumentParser(prog="observables_ANNNI_mps")
parser.add_argument("L", help="Number of spins", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h external transverse field", type=float
)
parser.add_argument(
    "h_f", help="Final value of h external transverse field", type=float
)
parser.add_argument(
    "k_i", help="Starting value of J2 next nearest interaction coefficient", type=float
)
parser.add_argument(
    "k_f", help="Final value of J2 next nearest interaction coefficient", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("o", help="Observable we want to compute. Available are 'wl', 'el', 'thooft'", type=str)
parser.add_argument("chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="ANNNI", type=str
)
parser.add_argument(
    "-mo", "--moment", help="Moment degree of the Free energy. E.g. Magnetization -> First Moment, Susceptibility -> Second Moment, etc. Available are 1,2,4", default=1, type=int
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval_h = np.linspace(args.h_i, args.h_f, args.npoints)
interval_k = np.linspace(args.k_i, args.k_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/2_ANNNI"
    path_tensor = "D:/code/projects/2_ANNNI"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm98/Desktop/projects/2_ANNNI"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm/Desktop/projects/2_ANNNI"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num) 


# define moment
if args.moment == 1:
    moment = "first"
if args.moment == 2:
    moment = "second"
if args.moment == 4:
    moment = "fourth"

# ---------------------------------------------------------
# Observables
# ---------------------------------------------------------
for chi in args.chis:
    M = []
    LM = []
    for h in interval_h:
        for k in interval_k:
            chain_mps = MPS(L=args.L, d=args.d, model=args.model, chi=chi, h=h, k=k)

            chain_mps.load_sites(path=path_tensor, precision=precision)
            
            if args.o == "mag":
                print(f"Magnetization for h:{h:.{precision}f}, k:{k:.{precision}f}")
                chain_mps.order_param()
                if args.moment == 1:
                    M.append(chain_mps.mpo_first_moment().real/chain_mps.L)
                elif args.moment == 2:
                    M.append(chain_mps.mpo_second_moment().real/(chain_mps.L**2))
                elif args.moment == 4:
                    M.append(chain_mps.mpo_fourth_moment().real/(chain_mps.L**4))
            
            elif args.o == "loc_mag":
                print(f"Local magnetization for h:{h:.{precision}f}, k:{k:.{precision}f}")
                chain_mps.local_param(site=(chain_mps.L // 2))
                LM.append(chain_mps.mpo_first_moment().real)
            else:
                raise ValueError("Select a valid observable. Available are 'mag', 'loc_mag'")

    if args.o == "mag":
        M = np.array(M)
        M = np.array_split(M, args.npoints)
        np.save(
                    f"{parent_path}/results/mag_data/magnetization_{moment}_moment_{args.model}_L_{args.L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}.npy",
                    M,
                )
    elif args.o == "loc_mag":
        LM = np.array(LM)
        LM = np.array_split(LM, args.npoints)
        np.save(
                    f"{parent_path}/results/mag_data/local_magnetization_{args.model}_L_{args.L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}.npy",
                    LM,
                )


import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision

parser = argparse.ArgumentParser(prog="get_rdms")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "hx_i", help="Starting value of x parameter in the phase space", type=float
)
parser.add_argument(
    "hx_f", help="Final value of x parameter in the phase space", type=float
)
parser.add_argument(
    "hy_i", help="Starting value of y parameter in the phase space", type=float
)
parser.add_argument(
    "hy_f", help="Final value of y parameter in the phase space", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Number of sites in the chain", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument(
    "-m", "--model", help="Model to simulate, available are 'Ising', 'Z2', 'ANNNI', 'Cluster', 'XXZ'. By default Cluster", default="Cluster", type=str
)
parser.add_argument(
    "-e", "--eps", help="Weight to the degeneracy breaking term", default=1e-5, type=float
)
parser.add_argument(
    "-w", "--which", help="Number of sites for the reduced density matrices. It will be centered in the middle of the chain", default=1, type=int
)
parser.add_argument(
    "-d", "--deg", help="Degeneracy method used during the DMRG. Depends on the model used", default=1, type=int
)
parser.add_argument(
    "-i",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)

args = parser.parse_args()

# define the physical dimension
d = int(2)

# define the interval of equally spaced values of external field
if args.interval == "lin":
    interval_x = np.linspace(args.hx_i, args.hx_f, args.npoints)
    interval_y = np.linspace(args.hy_i, args.hy_f, args.npoints)
    num = (args.hx_f - args.hx_i) / args.npoints
    precision = get_precision(num)
elif args.interval == "log":
    interval_x = np.logspace(args.hx_i, args.hx_f, args.npoints)
    interval_y = np.logspace(args.hy_i, args.hy_f, args.npoints)
    num = (interval_x[-1]-interval_x[0]) / args.npoints
    precision = get_precision(num)

if args.model == "Ising":
    model_path = "0_ISING"

elif args.model == "Z2":
    model_path = "1_Z2"

elif args.model == "ANNNI":
    model_path = "2_ANNNI"

elif args.model == "Cluster" or args.model == "Cluster-XY":
    model_path = "3_CLUSTER"

elif args.model == "XXZ":
    model_path = "4_XXZ"
else:
    raise SyntaxError("Model not valid. Choose among 'Ising', 'Z2', 'ANNNI', 'Cluster', 'Cluster-XY', 'XXZ'")
# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = f"G:/My Drive/projects/{model_path}"
    path_tensor = f"D:/code/projects/{model_path}"
    path_rdms = "C:/Users/HP/Desktop/fidelity-phase-tran"
elif args.path == "mac":
    parent_path = f"/Users/fradm98/Google Drive/My Drive/projects/{model_path}"
    path_tensor = f"/Users/fradm98/Desktop/projects/{model_path}"
    path_rdms = "/Users/fradm98/Desktop/fidelity-phase-tran"
elif args.path == "marcos":
    parent_path = f"/Users/fradm/Google Drive/My Drive/projects/{model_path}"
    path_tensor = f"/Users/fradm/Desktop/projects/{model_path}"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

print(f"Calculating rdms...")
print(f"{args.model} model selected")
for L in args.Ls:
    start = L // 2 - args.which // 2
    # Generate the list of subsystems using list comprehension
    sites = [start + i for i in range(args.which)]
    for chi in args.chis:  # L // 2 + 1
        print(f"L:{L}, chi:{chi}")
        rdms_chi = []
        for hy in interval_y:
            rdms_y = []
            for hx in interval_x:
                print(f"for hx:{hx:.{precision}f}, hy:{hy:.{precision}f}")
                if args.model == "Ising" or args.model == "Cluster":
                    chain = MPS(L=L, d=d, chi=chi, model=args.model, eps=args.eps, h=hx, J=hy)
                elif args.model == "ANNNI" or args.model == "XXZ":
                    chain = MPS(L=L, d=d, chi=chi, model=args.model, eps=args.eps, k=hx, h=hy, J=1)
                elif args.model == "Cluster-XY":
                    chain = MPS(L=L, d=d, chi=chi, model=args.model, eps=args.eps, lx=0, ly=hy, h=hx, J=1)
                chain.load_sites(path=path_tensor, precision=precision)
                rdm = chain.reduced_density_matrix(sites)
                rdms_y.append(rdm)
            rdms_chi.append(rdms_y)
        
        
        rdms_chi = np.asarray(rdms_chi).reshape((len(interval_x),len(interval_y),d**args.which,d**args.which))
        rdms_chi = rdms_chi[::-1]
        np.save(f"{parent_path}/results/rdms_data/{args.which}_sites-rdms_dmrg_{args.model}_L_{L}_hx_{args.hx_i}-{args.hx_f}_hy_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_degeneracy_method_{args.deg}_eps_{args.eps}_guess_path_chi_{chi}.npy", rdms_chi)


from scipy.sparse.linalg import LinearOperator
import numpy as np

class TensorMultiplierOperator(LinearOperator):
    r"""
    TensorMultiplierOperator

    This is a subclass of LinearOperator from PyLops. It serves
    as a new operator that implements the matrix vector product
    between the effective hamiltonian, given in tensor form by the environment
    left, right and the mpo in the middle - and the guess vector
    given by the flattened tensor in the site to optimize.
    The function implementing this is in the class MPS of qs-mps.

    """
    def __init__(self, shape, matvec, dtype=None):
        super(TensorMultiplierOperator, self).__init__(dtype=np.dtype(dtype), shape=shape)
        self.matvec = matvec

    def _matvec(self, x):
        return self.matvec(x)


from qs_mps.utils import *
from qs_mps.sparse_hamiltonians_and_operators import sparse_pauli_x, sparse_pauli_z
from qs_mps.lattice import Lattice
import numpy as np
from scipy.sparse import identity, csc_array, linalg
from ncon import ncon
import time

class H_Z2_gauss:
    def __init__(
        self,
        l,
        L,
        model: str,
        lamb: float=0,
        J: float=1,
        U: float=1e+3,
    ):
        self.L = L
        self.l = l
        self.model = model
        self.charges = np.ones((l, L))
        self.lamb = lamb
        self.J = J
        self.U = U
        self.latt = Lattice((self.L, self.l), (False, False))
        self.dof = self.latt.nlinks
        self.sector = self._define_sector()

    def add_charges(self, rows: list, columns: list):
        """
        add_charges

        This function adds the charges to the background
        vacuum sector (all positive charges). The number of
        charges we add are given by the len of each indices list

        rows: list - row indices of the charges to add
        columns: list - column indices of the charges to add
        """
        for i, j in zip(rows, columns):
            assert ((i != -1) and (j != -1)) or (
                (i == self.l - 1) and (j == self.L - 1)
            ), "Do not choose the last charge! We use it for Gauss Law constraint"
            self.charges[j, i] = -1

        self.charges = np.flip(self.charges, axis=0)
        return self

    def _define_sector(self):
        particles = 0
        for charge in self.charges.flatten():
            if charge == -1:
                particles += 1
        
        if particles == 0:
            sector = "vacuum_sector"
        else:
            sector = f"{particles}_particle(s)_sector"
            self.sector = sector
        return sector

    def local_term(self, link):
        sigma_x = sparse_pauli_x(n=link, L=self.latt.nlinks)
        return sigma_x

    def plaquette_term(self, loop):
        plaq = identity(n=2**self.latt.nlinks)
        for link in loop:
            plaq = plaq @ sparse_pauli_z(n=link, L=self.latt.nlinks)
        return plaq

    def gauge_constraint(self, site):
        links = self.latt.star(site=site, L=self.L, l=self.l)
        G = identity(n=2**self.latt.nlinks)
        filtered_links = [element for element in links if element != 0]
        # print("links:")
        # print(filtered_links)
        for link in filtered_links:
            G = G @ sparse_pauli_x(n=link - 1, L=self.latt.nlinks)

        return G

    def hamiltonian(self):
        loc = csc_array((2**self.latt.nlinks, 2**self.latt.nlinks))
        # local terms
        for link in range(self.latt.nlinks):
            loc += self.local_term(link)

        plaq = csc_array((2**self.latt.nlinks, 2**self.latt.nlinks))
        # plaquette terms
        for loop in self.latt.plaquettes(from_zero=True):
            plaq += self.plaquette_term(loop)

        # gauge constraint
        G = 0
        I = identity(n=2**self.latt.nlinks)
        for site in self.latt.sites:
            # print(site)
            g = self.gauge_constraint(site)
            G += (g - self.charges[site[1], site[0]] * I) @ (
                g - self.charges[site[1], site[0]] * I
            )
        return - (self.J * loc) - (self.lamb * plaq) + (self.U * G)

    def diagonalize(self, v0: np.ndarray=None, sparse: bool=True, save: bool=True, path: str=None, precision: int=2, spectrum: str="gs", cx: list=None, cy: list=None):
        H = self.hamiltonian()

        if sparse:
            if spectrum == "all":
                k = (2 ** len(self.latt.plaquettes()))
            elif spectrum == "gs":
                k = 1
            e, v = linalg.eigsh(H, k=k, which="SA", v0=v0)
        else:
            H = H.toarray()
            e, v = np.linalg.eigh(H)
        if save:
            # print(self.sector)
            np.save(path+f"/results/eigenvectors/ground_state_direct_lattice_{self.l-1}x{self.L-1}_{self.sector}_{cx}-{cy}_U_{self.U}_h_{self.lamb:.{precision}f}.npy", v[:,0])
        return e, v

    # observables
    def wilson_loop(self, psi, sites):
        loop = self.latt.plaquettes(from_zero=True)
        plaq = self.plaquette_term(loop[sites])
        exp_val_wilson_loop = np.real(psi.T @ plaq @ psi)
        return exp_val_wilson_loop
    
    def h_electric_field(self, psi):
        exp_val_h_links = []
        for link in range((self.l*(self.L-1))):
            obs = self.local_term(link)
            exp_val_electric_field = np.real(psi.T @ obs @ psi)
            exp_val_h_links.append(exp_val_electric_field)
        exp_val_h_links = np.array(exp_val_h_links).reshape((self.l,self.L-1))
        return exp_val_h_links
    
    def v_electric_field(self, psi):
        exp_val_v_links = []
        for link in range((self.l*(self.L-1)),self.dof):
            obs = self.local_term(link)
            exp_val_electric_field = np.real(psi.T @ obs @ psi)
            exp_val_v_links.append(exp_val_electric_field)
        exp_val_v_links = np.array(exp_val_v_links).reshape((self.l-1,self.L))
        return exp_val_v_links
    
    def electric_field(self, psi, E):
        E[::2,1::2] = self.h_electric_field(psi)
        E[1::2,::2] = self.v_electric_field(psi)
        return E

    def v_thooft_idx(self, plaq_tot_spl: np.ndarray, mpo_site: int, l: int):
        plaqs = [pl for pl in plaq_tot_spl[mpo_site]].copy()
        # plaqs.reverse()
        pauli = []
        for i in range(self.l-(l+1)):
            pauli = pauli + [plaqs[i][0]]

        return pauli
    
    def h_thooft_idx(self, plaq_tot_spl: np.ndarray, mpo_site: int, l: int):
        plaqs = np.swapaxes(plaq_tot_spl, axis1=0, axis2=1)
        plaqs = [pl for pl in plaqs].copy()
        plaqs.reverse()
        plaqs_h = plaqs[l].copy()
        pauli = []
        for i in range(mpo_site+1):
            pauli = pauli + [plaqs_h[i][3]]
        
        return pauli

    def thooft(self, psi: np.ndarray, mpo_site: int, l: int, direction: str):
        plaq_tot = self.latt.plaquettes(from_zero=True)
        plaq_tot_spl = np.array_split(plaq_tot, self.L-1)
        if direction == "vertical":
            pauli = self.v_thooft_idx(plaq_tot_spl=plaq_tot_spl, 
                                      mpo_site=mpo_site, 
                                      l=l)
        if direction == "horizontal":
            pauli = self.h_thooft_idx(plaq_tot_spl=plaq_tot_spl, 
                                      mpo_site=mpo_site, 
                                      l=l)
            
        op = identity(n=2**self.latt.nlinks)
        for idx in pauli:
            op = op @ sparse_pauli_x(n=idx, L=self.latt.nlinks)
        
        thooft_string = (psi.T @ op @ psi).real
        return thooft_string


# lamb = 0
# U = 1e+3
# Z2_exact = H_Z2_gauss(L=4, l=3, model="Z2_dual", lamb=lamb, U=U)
# Z2_exact.add_charges([0,2],[1,1])
# print(f"charges:\n{Z2_exact.charges}")
# print(f"degrees of freedom:\n{Z2_exact.dof}")
# print(f"lattice:\n{Z2_exact.latt._lattice_drawer.draw_lattice()}")
# e, v = Z2_exact.diagonalize(save=False, sparse=False)
# psi = v[:,0]
# print(psi)
# print(f"spectrum:\n{e}")
# # print(f"Delta Energy gs - ex: {e[0]-e[1]}\nth: {8*lamb}")
# # print(f"H:\n{H.toarray()}")
# print(f"ground state:\n{v[:,0]}")
# X = sparse_pauli_z(n=0,L=4) @ sparse_pauli_z(n=1,L=4) @ sparse_pauli_z(n=2,L=4) @ sparse_pauli_z(n=3,L=4) 

# exp_val = (psi.T @ X @ psi).real
# print(exp_val)


# from qs_mps.mpo_class import MPO_ladder
# exp_val = []
# en = []
# en_1 = []
# en_mpo = []
# en_mpo_1 = []
# # delta_e = []
# hs = list(np.arange(-6, 6, 0.1))
# hs.reverse()
# hs.pop()
# l = 2
# L = 2
# dof = (2*l*L - l - L)
# v0 = np.array([-0.25 for _ in range(2**dof)])
# for h in hs:
#     Z2_mpo = MPO_ladder(L=L, l=l-1, model="Z2", lamb=h)
#     e, v = Z2_mpo.diagonalize()
#     en_mpo.append(e[0])
#     en_mpo_1.append(e[1])
#     Z2_exact = H_Z2_gauss(L=L, l=l, model="Z2", lamb=h, U=1e+3)
#     H, e, v = Z2_exact.diagonalize(v0=v0)
#     print(e)
#     en.append(e[0])
#     en_1.append(e[1])
#     # delta_e.append(np.abs(e[1]-e[0]))
#     # print(f"spectrum:\n{e}")
#     # print(f"{Z2_exact.latt.plaquettes()[1]}")
#     # loop = Z2_exact.latt.plaquettes(from_zero=True)[1]
#     # loop = Z2_exact.plaquette_term(loop=loop)
#     psi = v[:,0]
#     # # exp = ncon([psi.T,loop.toarray(),psi],[[1],[1,2],[2]]).real
#     # exp = (psi.T @ loop @ psi).real
#     # exp_val.append(exp)
#     v0 = psi

# print(f"lattice:\n{Z2_exact.latt._lattice_drawer.draw_lattice()}")
# # plt.plot(hs, np.abs(exp_val))
# # plt.show()
# plt.plot(hs, en_mpo, 'o', color='g')
# plt.plot(hs, en_mpo_1, '-', color='g')
# plt.plot(hs, en, '--', color='r')
# plt.plot(hs, en_1, '-', color='r')
# plt.show()
# # plt.plot(hs, 8*np.asarray(hs))
# # plt.show()


import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision, save_list_of_lists, access_txt, swap_rows
from qs_mps.gs_multiprocessing import ground_state_ising

import signal
import time

# Define a function to handle the timeout
def timeout_handler(signum, frame):
    raise TimeoutError("Algorithm took too long to execute")

# Set the signal handler
signal.signal(signal.SIGALRM, timeout_handler)

# DENSITY MATRIX RENORMALIZATION GROUP to find ground states of the 
# 1D Ising Transverse Field model changing the transverse field parameters

parser = argparse.ArgumentParser(prog="gs_search_Ising")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Spin chain lengths", nargs="*", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d","--dimension", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="trapezoidal", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=8,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-10,
    type=float,
)
parser.add_argument(
    "-fid",
    "--fidelity",
    help="Fidelity with exact solution. Doable only for small Ising chains",
    action="store_true",
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-tr",
    "--training",
    help="Save all the energies during the variational optimization. By default True",
    action="store_false",
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval_hx = np.linspace(args.h_i, args.h_f, args.npoints).tolist()
interval_hy = np.linspace(args.h_i, args.h_f, args.npoints)
interval_hx.reverse()

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/0_ISING"
    path_tensor = "D:/code/projects/0_ISING"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm98/Desktop/projects/0_ISING"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm/Desktop/projects/0_ISING"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

path = f"{parent_path}/results/tensors"
num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num)


# ---------------------------------------------------------
# DMRG
# ---------------------------------------------------------
for L in args.Ls:
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False
    for chi in args.chis:  # L // 2 + 1
        # init_state = np.zeros((1, 2, 1))
        # init_state[0, 0, 0] = 1
        # init_tensor = [init_state for _ in range(L)]
        energy_chi = []
        entropy_chi = []
        schmidt_vals_chi = []
        time_chi = []
        exceptions_chi = np.zeros((len(interval_hy),len(interval_hx)))
        for idx0, J in enumerate(interval_hy):
            energy_J = []
            entropy_J = []
            schmidt_vals_J = []
            t_slice = []
            new_timeout_secs = 5
            for idx1, h in enumerate(interval_hx):
                chain = MPS(
                    L=L,
                    d=2,
                    model="Ising",
                    chi=chi,
                    h=h,
                    J=J,
                    eps=1e-5,
                )                
                if h == interval_hx[0]:
                    # init_tensor = [init_state for _ in range(L)]
                    # chain.sites = init_tensor.copy()
                    # chain.enlarge_chi(type_shape="rectangular", prnt=False)
                    chain._random_state(3, chi=chi)
                    chain.canonical_form(trunc_chi=False, trunc_tol=True)
                else:
                    chain.sites = init_tensor.copy()

                # Set the timeout period (in seconds)
                timeout_secs = new_timeout_secs # You can change this value according to your requirement

                # Set the alarm
                signal.alarm(int(timeout_secs+1))
                print(f"New timeout seconds: {int(timeout_secs+1)}")
                try:
                    # Call your algorithm function with the initial parameter
                    print("Running with guess state:")
                    energy, entropy, schmidt_vals, t_dmrg = chain.DMRG(
                        trunc_tol=False,
                        trunc_chi=True,
                        where=args.where,
                        bond=args.bond,
                    )
                except TimeoutError as e:
                    print(e)
                    # Modify the parameter and call the algorithm again
                    chain._random_state(seed=7, type_shape="rectangular", chi=chi)
                    chain.canonical_form()
                    print("Running with random state:")
                    energy, entropy, schmidt_vals, t_dmrg = chain.DMRG(
                        trunc_tol=False,
                        trunc_chi=True,
                        where=args.where,
                        bond=args.bond,
                    )
                    exceptions_chi[idx0,idx1] = 1
                else:
                    # Cancel the alarm if the algorithm finishes before the timeout
                    signal.alarm(0)

                t_slice.append(t_dmrg)
                t_mean = np.mean(t_slice)
                t_std = np.std(t_slice)
                new_timeout_secs = t_mean + 3*t_std
                
                print(f"energy of h:{h:.{precision}f}, J:{J:.{precision}f} is:\n {energy}")
                print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")

                chain.save_sites(path=path_tensor, precision=precision)
                if args.training:
                    energy_J.append(energy)
                else:
                    energy_J.append(energy[-1])
                entropy_J.append(entropy)
                schmidt_vals_J.append(schmidt_vals)

                if h == interval_hx[0]:
                    init_tensor_J = chain.sites.copy()
                init_tensor = chain.sites.copy()
            
            init_tensor = init_tensor_J.copy()

            time_chi.append(t_slice)
            energy_chi.append(energy_J)
            entropy_chi.append(entropy_J)
            schmidt_vals_chi.append(schmidt_vals_J)
        # energy_chi, entropy_chi, schmidt_vals_chi = ground_state_ising(
        #     args_mps=args_mps, multpr=False, param=interval
        # )

        if args.bond == False:
            args.where = "all"

        np.save(f"{parent_path}/results/energy_data/", exceptions_chi)

        if args.training:
            energy_chi = np.swapaxes(swap_rows(np.asarray(energy_chi)), axis1=0, axis2=1)
            np.save(
                f"{parent_path}/results/energy_data/energies_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_J_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
            energy_min = []
            for i in range(len(interval_hx)):
                energy_min_i = []
                for j in range(len(interval_hy)):
                    energy_min_i.append(energy_chi[i][j][-1])
                energy_min.append(energy_min_i)
            np.save(f"{parent_path}/results/energy_data/energy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_J_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}", energy_min)
        else:
            energy_chi = np.array(energy_chi)
            energy_chi = np.array_split(energy_chi, args.npoints)
            save_list_of_lists(
                f"{parent_path}/results/energy_data/energies_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
        entropy_chi = np.array(entropy_chi)
        entropy_chi = np.array_split(entropy_chi, args.npoints)
        save_list_of_lists(
            f"{parent_path}/results/entropy/{args.where}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
            entropy_chi,
        )
        schmidt_vals_chi = np.array(schmidt_vals_chi)
        schmidt_vals_chi = np.array_split(schmidt_vals_chi, args.npoints)
        save_list_of_lists(
            f"{parent_path}/results/entropy/{args.where}_schmidt_values_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
            schmidt_vals_chi,
        )
        if args.where == "all":
            entropy_mid = access_txt(
                f"{parent_path}/results/entropy/{args.where}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                L // 2,
            )
            np.savetxt(
                f"{parent_path}/results/entropy/{L // 2}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                entropy_mid,
            )



# import packages
from qs_mps.mps_class import MPS
from qs_mps.utils import *
import matplotlib.pyplot as plt
from ncon import ncon
import scipy
from scipy.sparse import csr_array
import time
import argparse

# variational compression changing bond dimension
# parameters

parser = argparse.ArgumentParser(prog="Time Evolution")
parser.add_argument("L", help="Spin chain length", type=int)
parser.add_argument("chi", help="Simulated bond dimension", type=int)
parser.add_argument(
    "h_ev", help="It will give you the magnitude of the quench", type=float
)
parser.add_argument(
    "trotter_steps",
    help="Different trotter steps, changes the precision of the evolution",
    nargs="+",
    type=int,
)
parser.add_argument(
    "-f", "--flip", help="Flip the middle site or not", action="store_true"
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-t", "--time", help="Final time of the evolution", default=10, type=float
)
parser.add_argument(
    "-h_ti",
    "--h_transverse_init",
    help="Initial transverse field before the quench",
    default=0,
    type=float,
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=8,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-10,
    type=float,
)
parser.add_argument(
    "-fid",
    "--fidelity",
    help="Fidelity with exact solution. Doable only for small Ising chains",
    action="store_true",
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, "
    +"should be between 1 and (L-1). Default -1 to get the L//2 middle bond and could assume value -2 for all bonds schmidt values",
    default=-1,
    type=int,
)
args = parser.parse_args()
delta = args.time / args.trotter_steps
if args.where == -1:
    args.where = args.L // 2
<<<<<<< HEAD:src/mps/variational_compression_different_chi.py
elif args.where == -2:
    args.bond = False
=======
    
>>>>>>> origin/main:src/qs_mps/applications/0_ISING/TEBD_different_delta.py
# ---------------------------------------------------------
# variational truncation mps
for trotter_step in args.trotter_steps:  # L // 2 + 1
    delta = args.time / trotter_step
    chain = MPS(
        L=args.L,
        d=2,
        model=args.model,
        chi=args.chi,
        h=args.h_transverse_init,
        eps=0,
        J=1,
    )
    chain._random_state(seed=3, chi=args.chi)
    chain.canonical_form(trunc_chi=False, trunc_tol=True)
    # chain.sweeping(trunc_chi=False, trunc_tol=True, n_sweeps=2)
    init_state = np.zeros((1, 2, 1))
    init_state[0, 0, 0] = 1
    for i in range(chain.L):
        chain.sites[i] = init_state
    (
        mag_mps_tot,
        mag_mps_loc_X,
        mag_mps_loc,
        overlap,
        errors,
        schmidt_values,
    ) = chain.TEBD_variational(
        trotter_steps=args.trotter_steps,
        delta=delta,
        h_ev=args.h_ev,
        flip=args.flip,
        n_sweeps=args.number_sweeps,
        conv_tol=args.conv_tol,
        fidelity=args.fidelity,
        bond=args.bond,
        where=args.where,
    )

    if args.bond == False:
        args.where = "all"

    np.savetxt(
<<<<<<< HEAD:src/mps/variational_compression_different_chi.py
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_tot_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        mag_mps_tot,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_X_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        mag_mps_loc_X,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        mag_mps_loc,
    )
    mag_mps_loc_Z = access_txt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        args.L // 2,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_Z_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        mag_mps_loc_Z,
    )
    # np.savetxt(
    #     f"/data/fdimarca/projects/0_ISING/results/fidelity_data/fidelity_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}", overlap
    # )
    save_list_of_lists(
        f"/data/fdimarca/projects/0_ISING/results/errors_data/errors_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
        errors,
    )
    save_list_of_lists(
        f"/data/fdimarca/projects/0_ISING/results/bonds_data/{args.where}_bond_schmidt_values_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{chi}",
=======
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_tot_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        mag_mps_tot,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_X_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        mag_mps_loc_X,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        mag_mps_loc,
    )
    mag_mps_loc_Z = access_txt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        args.L // 2,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/mag_data/mag_mps_loc_Z_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        mag_mps_loc_Z,
    )
    # np.savetxt(
    #     f"/data/fdimarca/projects/0_ISING/results/fidelity_data/fidelity_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}", overlap
    # )
    save_list_of_lists(
        f"/data/fdimarca/projects/0_ISING/results/errors_data/errors_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
        errors,
    )
    np.savetxt(
        f"/data/fdimarca/projects/0_ISING/results/bonds_data/{args.where}_bond_schmidt_values_{args.model}_L_{args.L}_flip_{args.flip}_delta_{delta}_chi_{args.chi}",
>>>>>>> origin/main:src/qs_mps/applications/0_ISING/TEBD_different_delta.py
        schmidt_values,
    )



import matplotlib.pyplot as plt
import numpy as np
from qs_mps.utils import plot_results_DMRG, plot_colormaps_evolution, get_precision
import argparse

# default parameters of the plot layout
plt.rcParams["text.usetex"] = True  # use latex
plt.rcParams["font.size"] = 13
plt.rcParams["figure.dpi"] = 300
plt.rcParams["figure.constrained_layout.use"] = True


parser = argparse.ArgumentParser(prog="Plot Ground States")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values or trotter steps",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field)", type=float
)
parser.add_argument(
    "loc",
    help="From which computer you want to access the drive. Useful for the path spec. Available are: 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument(
    "obs",
    help="Results we want to plot. Available are: 'energy', 'error', 'entropy', 'wilson_loop', 'thooft', 'mag', 'energy_tr', 'error_tr', 'entropy_tot'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Number of rungs per ladder", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument("-lx", "--sites", help="Number of sites in the wilson loop", nargs="*", type=int)
parser.add_argument("-ly", "--ladders", help="Number of ladders in the wilson loop", nargs="*", type=int)
parser.add_argument(
    "-f", "--h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "-del", "--delta", help="Width of each time slice during the time evolution. Should be 'small enough'", type=float
)
parser.add_argument(
    "-ev", "--h_ev", help="Quench value of h (external transverse field on the dual lattice)", type=float
)

parser.add_argument(
    "-dir", "--direction", help="Direction of the string", default="hor", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
)
parser.add_argument(
    "-mo", "--moment", help="Moment of the order parameter. Available are 1,2, and 4. By default 1", default=1, type=int
)
parser.add_argument(
    "-t", "--time", help="Plot observables of time evolution", action="store_true"
)
parser.add_argument(
    "-v",
    "--save",
    help="Save plot, by default True",
    action="store_false",
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-k", "--marker", help="Marker to use for the Plot", default="+", type=str
)
parser.add_argument("-z", "--m_size", help="marker size", default=20, type=int)
parser.add_argument(
    "-li",
    "--linewidth",
    help="Linewidth of the marker",
    default=1,
    type=float,
)
parser.add_argument(
    "-a",
    "--alpha",
    help="Alpha of the marker",
    default=1,
    type=float,
)
parser.add_argument(
    "-n",
    "--n_points",
    help="Percentage of number of points we want to plot, from 0 to 1",
    default=1,
    type=float,
)
parser.add_argument(
    "-c", "--cmap", help="colormap for the plots", default="viridis", type=str
)
parser.add_argument(
    "-i",
    "--interpolation",
    help="interpolation for the colormap plots",
    default="antialiased",
    type=str,
)
parser.add_argument(
    "-int",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)
parser.add_argument(
    "-d",
    "--dim",
    help="Dimension to plot, by defalut plots two dimensional plots. If True 3D plots, by default False",
    action="store_true",
)
parser.add_argument(
    "-e",
    "--exact",
    help="Exact results to compare the mps ones. If True we compare - be default False",
    action="store_true",
)

args = parser.parse_args()

# define the interval of equally spaced values of external field or trotter steps
if args.time:
    interval = range(args.npoints+1)
    precision = get_precision(args.h_i)
    if args.interval == "lin":
        xscale = "linear"
    elif args.interval == "log":
        xscale = "log"
else:
    if args.interval == "lin":
        interval = np.linspace(args.h_i, args.h_f, args.npoints)
        xscale = "linear"
    elif args.interval == "log":
        interval = np.logspace(args.h_i, args.h_f, args.npoints)
        xscale = "log"

    # define the precision
    num = (args.h_f - args.h_i) / args.npoints
    precision = get_precision(num)

# define the direction
if args.direction == "ver":
    direction = "vertical"
elif args.direction == "hor":
    direction = "horizontal"  


# define the moment
if args.moment == 1:
    moment = "first"
elif args.moment == 2:
    moment = "second"
elif args.moment == 4:
    moment = "fourth"


plot_val = [
    "energy",
    "error",
    "entropy",
    "wilson_loop",
    "thooft",
    "mag"
]
plot_cmap = ["energy_tr", "error_tr", "entropy_tot"]

if args.loc == "pc":
    parent_path = "G:/My Drive/projects/1_Z2/"
elif args.loc == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2/"
elif args.loc == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2/"
else:
    raise SyntaxError("insert a valid location: 'pc', 'mac', 'marcos'")

yscale = "linear"

for L in args.Ls:
    # define the sector by looking of the given charges
    if len(args.charges_x) == 0:
        sector = "vacuum_sector"
        charges_x = None
        charges_y = None
    else:
        sector = f"{len(args.charges_x)}_particle(s)_sector"
        charges_x = args.charges_x
        charges_y = args.charges_y

    # where to look at for the entropy
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = "all"


    if args.obs == "energy":
        title = f"Energy: lattice = ${args.l}$x${L-1}$ ; {sector}"
        fname_obs = "energies"
        fname_ex_obs = "energies"
        path = parent_path + f"results/energy_data"
        path_ex = parent_path + f"results/exact/energy_data"
        path_save = parent_path + f"figures/energy/"
        ylabel = "$\\langle E\\rangle$"
        txt = True

    elif args.obs == "error":
        title = f"Error: lattice = ${args.l}$x${L-1}$ ; {sector}"
        if args.time:
            fname_obs = "errors_quench_dynamics"
            fname_ex_obs = "errors_quench_dynamics"
        else:
            fname_obs = "errors"
            fname_ex_obs = "errors"
        path = parent_path + f"results/error_data"
        path_ex = parent_path + f"results/exact/error_data"
        path_save = parent_path + f"figures/error/"
        ylabel = "$\\left|\\left|\\left| \psi\\right>- \\left| \hat{O}\phi\\right>\\right|\\right|^2$"
        yscale = "log"
        txt = True

    elif args.obs == "entropy":
        title = (
            f" ${args.where}-th$ Bond Entanglement Entropy: lattice = ${args.l}$x${L-1}$ ; {sector}"
            # + f" $h \in ({args.h_i},{args.h_f})$"
        )
        if args.time:
            fname_obs = f"{args.where}_bond_entropy_quench_dynamics"
            fname_ex_obs = f"{args.where}_bond_entropy_quench_dynamics"
        else:
            fname_obs = f"{args.where}_bond_entropy"
            fname_ex_obs = f"{args.where}_bond_entropy"

        path = parent_path + f"results/entropy_data"
        path_ex = parent_path + f"results/exact/entropy_data"
        path_save = parent_path + f"figures/entropy/"
        ylabel = "entanglement von neumann entropy $(S_{\chi})$"
        txt = True

    elif args.obs == "wilson_loop":
        title = f"Wilson Loop: lattice = ${args.l}$x${L-1}$ ; {sector}"
        fname_obs = f"wilson_loop_{moment}_moment"
        fname_ex_obs = f"wilson_loop_{moment}_moment"
        path = parent_path + f"results/wilson_loops"
        path_ex = parent_path + f"results/exact/wilson_loops"
        path_save = parent_path + f"figures/wilson_loops/"
        ylabel = "$\\langle W\\rangle$"
        txt = False

    elif args.obs == "thooft":
        title = f"'t Hooft {direction} string: lattice = ${args.l}$x${L-1}$ ; {sector}" + " $\mu^z$ for " + f"L={args.sites[0]}, l={args.ladders[0]}" + " dual"
        fname_obs = f"thooft_string_{moment}_moment_{args.sites[0]}-{args.ladders[0]}_{direction}"
        fname_ex_obs = f"thooft_string_{moment}_moment_{args.sites[0]}-{args.ladders[0]}_{direction}"
        path = parent_path + f"results/thooft"
        path_ex = parent_path + f"results/exact/thooft"
        path_save = parent_path + f"figures/thooft/"
        ylabel = "$\\langle \sigma^x \dots \sigma^x \\rangle$"
        txt = False
    
    elif args.obs == "mag":
        title = f"Dual Magnetization: lattice = ${args.l}$x${L-1}$ ; {sector}" 
        fname_obs = f"dual_mag_{moment}_moment"
        fname_ex_obs = f"dual_mag_{moment}_moment"
        path = parent_path + f"results/mag_data"
        path_ex = parent_path + f"results/exact/mag_data"
        path_save = parent_path + f"figures/magnetization/"
        ylabel = "$\\langle \sigma^x \dots \sigma^x \\rangle$"
        txt = False

    elif args.obs == "entropy_tot":
        title = (
            f"All Bonds Entanglement Entropy: $lattice = {args.l}x{L}$ ; {sector}"
            + f" $h \in ({args.h_i},{args.h_f})$"
        )
        fname_obs = f"{args.where}_bond_entropy"
        path = parent_path + f"results/entropy_data"
        path_save = parent_path + f"figures/entropy/"
        xlabel = "bonds"
        x_ticks = range(args.L - 1)
        labels = range(1, args.L - 1 + 1)
        steps = len(x_ticks) // 5
        xticks = x_ticks[::steps]
        steps = len(labels) // 5
        xlabels = labels[::steps]
        y_ticks = np.arange(args.trotter_steps + 1)
        # labels = delta * y_ticks
        steps = len(y_ticks) // 5
        yticks = y_ticks[::steps]
        steps = len(labels) // 5
        ylabels = labels[::steps]
        X, Y = np.meshgrid(x_ticks, y_ticks)
        view_init = False

    else:
        raise SyntaxError(
            "insert a valid result to plot: 'energy', 'error', 'entropy', 'wilson_loop', 'thooft', 'mag', 'energy_tr', 'error_tr', 'entropy_tot'"
        )

    if args.obs in plot_val:
        if args.time:
            fname = f"{fname_obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}"
            fname_ex = f"{fname_ex_obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}"
            fname_save = f"{args.obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}"
        else:
            fname = f"{fname_obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}"
            fname_ex = f"{fname_ex_obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}"
            fname_save = f"{args.obs}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}"
        plot_results_DMRG(
            title,
            for_array=args.chis,
            interval=interval,
            fname=fname,
            path=path,
            fname_save=fname_save,
            path_save=path_save,
            ylabel=ylabel,
            yscale=yscale,
            xscale=xscale,
            save=args.save,
            exact=args.exact,
            fname_ex=fname_ex,
            path_ex=path_ex,
            delta=args.delta,
            time=args.time,
            marker=args.marker,
            m_size=args.m_size,
            linewidth=args.linewidth,
            alpha=args.alpha,
            n_points=args.n_points,
            cmap=args.cmap,
            txt=txt,
            precision=precision
        )

    if args.obs in plot_cmap:
        for chi in args.chis:
            title_fin = title + f" ; $\chi = {chi}$"
            fname = f"{fname_obs}_{args.model}_L_{args.L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}"
            fname_save = f"{args.obs}_{args.model}_L_{args.L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}"
            plot_colormaps_evolution(
                title=title_fin,
                fname=fname,
                path=path,
                fname_save=fname_save,
                path_save=path_save,
                xlabel=xlabel,
                xticks=xticks,
                xlabels=xlabels,
                yticks=yticks,
                ylabels=ylabels,
                X=X,
                Y=Y,
                save=args.save,
                cmap=args.cmap,
                interpolation=args.interpolation,
                d=args.dim,
                view_init=view_init,
            )



# import packages
import argparse
from qs_mps.mps_class import MPS
from qs_mps.utils import *

parser = argparse.ArgumentParser(prog="observables_Z2_mps")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-o", "--obs", help="Observable we want to compute. Available are 'wl', 'wl_av', 'el', 'thooft', 'mag', 'corr'", nargs="+", type=str)
parser.add_argument("-L", "--Ls", help="Number of rungs per ladder", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument("-lx", "--sites", help="coordinates of sites", nargs="*", type=int)
parser.add_argument("-ly", "--ladders", help="coordinates of ladders", nargs="*", type=int)
parser.add_argument(
    "-d", "--direction", help="Direction of the string", default="hor", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
)
parser.add_argument(
    "-mo", "--moment", help="Moment degree of the Free energy. E.g. Magnetization -> First Moment, Susceptibility -> Second Moment, etc. Available are 1,2,4", default=1, type=int
)
parser.add_argument(
    "-i",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)

args = parser.parse_args()

# define the physical dimension
d = int(2**(args.l))

# define the interval of equally spaced values of external field
if args.interval == "lin":
    interval = np.linspace(args.h_i, args.h_f, args.npoints)
    num = (interval[-1] - interval[0]) / args.npoints
    precision = get_precision(num)
elif args.interval == "log":
    interval = np.logspace(args.h_i, args.h_f, args.npoints)
    precision = int(np.max([np.abs(args.h_f),np.abs(args.h_i)]))

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_tensor = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")


# define the direction
if args.direction == "ver":
    direction = "vertical"
elif args.direction == "hor":
    direction = "horizontal"    

# define moment
if args.moment == 1:
    moment = "first"
if args.moment == 2:
    moment = "second"
if args.moment == 4:
    moment = "fourth"


# ---------------------------------------------------------
# Observables
# ---------------------------------------------------------
for L in args.Ls:
    # define the sector by looking of the given charges
    if len(args.charges_x) == 0:
        sector = "vacuum_sector"
        charges_x = None
        charges_y = None
    else:
        sector = f"{len(args.charges_x)}_particle(s)_sector"
        charges_x = args.charges_x
        charges_y = args.charges_y

    for chi in args.chis:
        W = []
        W_av = []
        E = []
        S = []
        M = []
        C = []
        for h in interval:
            lattice_mps = MPS(L=L, d=d, model=args.model, chi=chi, h=h)
            lattice_mps.L = lattice_mps.L - 1

            lattice_mps.load_sites(path=path_tensor, precision=precision, cx=charges_x, cy=charges_y)
            if sector != "vacuum_sector":
                lattice_mps.Z2.add_charges(charges_x, charges_y)
            
            if "wl" in args.obs:
                print(f"wilson loop for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                lattice_mps.Z2.wilson_Z2_dual(mpo_sites=args.sites, ls=args.ladders) #list(range(s))
                lattice_mps.w = lattice_mps.Z2.mpo.copy()
                if args.moment == 1:
                    # print(lattice_mps.mpo_first_moment().real)
                    W.append(lattice_mps.mpo_first_moment().real)
                elif args.moment == 2:
                    # print(lattice_mps.mpo_second_moment().real)
                    W.append(lattice_mps.mpo_second_moment().real)
                elif args.moment == 4:
                    # print(lattice_mps.mpo_fourth_moment().real)
                    W.append(lattice_mps.mpo_fourth_moment().real)

            if "wl_av" in args.obs:
                print(f"wilson loop average for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                wav = []
                for Ly in range(args.l):
                    for Lx in range(L-1):
                        lattice_mps.Z2.wilson_Z2_dual(mpo_sites=[Lx], ls=[Ly]) #list(range(s))
                        lattice_mps.w = lattice_mps.Z2.mpo.copy()
                        if args.moment == 1:
                            # print(lattice_mps.mpo_first_moment().real)
                            wav.append(lattice_mps.mpo_first_moment().real)
                        elif args.moment == 2:
                            # print(lattice_mps.mpo_second_moment().real)
                            wav.append(lattice_mps.mpo_second_moment().real)
                        elif args.moment == 4:
                            # print(lattice_mps.mpo_fourth_moment().real)
                            wav.append(lattice_mps.mpo_fourth_moment().real)
                N = len(wav)
                wav = 1/N*np.sum(wav)
                W_av.append(wav)

            if "el" in args.obs:
                print(f"electric field for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                E_h = np.zeros((2*args.l+1,2*L-1))
                E_h[:] = np.nan
                E_h = lattice_mps.electric_field_Z2(E_h)
                E.append(E_h)
            
            if "thooft" in args.obs:
                print(f"'t Hooft string for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                lattice_mps.Z2.thooft(site=args.sites, l=args.ladders, direction=direction)
                lattice_mps.w = lattice_mps.Z2.mpo.copy()
                S.append(lattice_mps.mpo_first_moment().real)

            if "mag" in args.obs:
                print(f"Magnetization for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                lattice_mps.order_param()
                if args.moment == 1:
                    #  print(lattice_mps.mpo_first_moment().real, (len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l))))
                    M.append(lattice_mps.mpo_first_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l))))
                elif args.moment == 2:
                    M.append(lattice_mps.mpo_second_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**2)
                elif args.moment == 4:
                    M.append(lattice_mps.mpo_fourth_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**4)

            if "corr" in args.obs:
                print(f"Correlator for h:{h:.{precision}f}, direct lattice lxL:{args.l}x{L-1}, chi:{chi}")
                c = lattice_mps.connected_correlator(site=args.sites[0], lad=args.ladders[0])
                C.append(c)


        if "wl" in args.obs:
            np.save(
                        f"{parent_path}/results/wilson_loops/wilson_loop_{moment}_moment_{args.sites}-{args.ladders}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        W,
                    )
        if "wl_av" in args.obs:
            np.save(
                        f"{parent_path}/results/wilson_loops/wilson_loop_average_{moment}_moment_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        W_av,
                    )
        if "el" in args.obs:
            np.save(
                        f"{parent_path}/results/electric_field/electric_field_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        E,
                    )
        if "thooft" in args.obs:
            np.save(
                        f"{parent_path}/results/thooft/thooft_string_{moment}_moment_{args.sites[0]}-{args.ladders[0]}_{direction}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        S,
                    )
        if "mag" in args.obs:
            np.save(
                        f"{parent_path}/results/mag_data/dual_mag_{moment}_moment_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        M,
                    )
        if "corr" in args.obs:
            C = np.array_split(C, args.npoints)
            np.save(
                        f"{parent_path}/results/mag_data/connected_correlator_s_{args.sites[0]}_l_{args.ladders[0]}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                        C,
                    )



import matplotlib.pyplot as plt
import numpy as np
from qs_mps.utils import plot_results_TEBD, plot_colormaps_evolution
import argparse

# default parameters of the plot layout
plt.rcParams["text.usetex"] = True  # use latex
plt.rcParams["font.size"] = 13
plt.rcParams["figure.dpi"] = 300
plt.rcParams["figure.constrained_layout.use"] = True


parser = argparse.ArgumentParser(prog="Plot Time Evolution")
parser.add_argument("L", help="Spin chain length", type=int)
parser.add_argument(
    "trotter_steps",
    help="It will give you the accuracy of the trotter evolution for fixed t",
    type=int,
)
parser.add_argument(
    "h_ev", help="It will give you the magnitude of the quench", type=float
)
parser.add_argument(
    "what",
    help="Results we want to plot. Available are: 'mag_tot', 'mag_z', 'mag_x', 'entropy', 'err_mag_tot', 'err_mag_z', 'err_mag_x', 'err_entropy', 'entropy_tot', 'mag_loc' ",
    type=str,
)
parser.add_argument(
    "loc",
    help="From which computer you want to access the drive. Useful for the path spec. Available are: 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument(
    "-f",
    "--flip",
    help="Flip the middle site or not. By defalut True",
    action="store_false",
)
parser.add_argument(
    "-q",
    "--quench",
    help="Type of quench. Available are 'flip', 'global'",
    default="global",
    type=str,
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-t", "--time", help="Final time of the evolution", default=10, type=float
)
parser.add_argument(
    "-s",
    "--save",
    help="Save plot, by default True",
    action="store_false",
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-k", "--marker", help="Marker to use for the Plot", default="+", type=str
)
parser.add_argument("-z", "--m_size", help="marker size", default=20, type=int)
parser.add_argument(
    "-l",
    "--linewidth",
    help="Linewidth of the marker",
    default=1,
    type=float,
)
parser.add_argument(
    "-a",
    "--alpha",
    help="Alpha of the marker",
    default=1,
    type=float,
)
parser.add_argument(
    "-n",
    "--n_points",
    help="Percentage of number of points we want to plot, from 0 to 1",
    default=1,
    type=float,
)
parser.add_argument(
    "-c", "--cmap", help="colormap for the plots", default="viridis", type=str
)
parser.add_argument(
    "-i",
    "--interpolation",
    help="interpolation for the colormap plots",
    default="antialiased",
    type=str,
)
parser.add_argument(
    "-d",
    "--dim",
    help="Dimension to plot, by defalut plots two dimensional plots. If True 3D plots, by default False",
    action="store_true",
)
parser.add_argument(
    "-e",
    "--exact",
    help="Exact results to compare the mps ones. If True we compare - be default False",
    action="store_true",
)

args = parser.parse_args()
if args.where == -1:
    args.where = args.L // 2
elif args.where == -2:
    args.bond = False
    args.where = "all"

# interval = np.linspace(args.h_i, args.h_f, args.npoints).tolist()
delta = args.time / args.trotter_steps

plot_val = [
    "mag_tot",
    "mag_z",
    "mag_x",
    "entropy",
    "err_mag_tot",
    "err_mag_z",
    "err_mag_x",
    "err_entropy",
]
plot_cmap = ["mag_loc", "entropy_tot"]

if args.loc == "pc":
    path_computer = "G:/My Drive/projects/0_ISING/"
elif args.loc == "mac":
    path_computer = "/Users/fradm98/Google Drive/My Drive/projects/0_ISING/"
elif args.loc == "marcos":
    path_computer = "/Users/fradm/Google Drive/My Drive/projects/0_ISING/"
else:
    raise SyntaxError("insert a valid location: 'pc', 'mac', 'marcos'")

if args.what == "mag_tot":
    title = f"Total Magnetization: $L = {args.L}$ ;" + " $h_{ev}=$ " + f"${args.h_ev}$"
    fname_what = "mag_mps_tot"
    fname_ex_what = "mag_exact_tot"
    path = path_computer + f"results/mag_data"
    path_ex = path_computer + f"results/exact/mag_data"
    path_save = path_computer + f"figures/magnetization/"
    ylabel = "$\\sum_{i=1}^L \\sigma_i^z$"

elif args.what == "mag_z":
    title = (
        f"Magnetization Z in the middle site: $L = {args.L}$ ;"
        + " $h_{ev}=$ "
        + f"${args.h_ev}$"
    )
    fname_what = "mag_mps_loc_Z"
    fname_ex_what = "mag_exact_loc_Z"
    path = path_computer + f"results/mag_data"
    path_ex = path_computer + f"results/exact/mag_data"
    path_save = path_computer + f"figures/magnetization/"
    ylabel = "$\\sigma_{L/2}^z$"

elif args.what == "mag_x":
    title = (
        f"Magnetization X in the middle site: $L = {args.L}$ ;"
        + " $h_{ev}=$ "
        + f"${args.h_ev}$"
    )
    fname_what = "mag_mps_loc_X"
    fname_ex_what = "mag_exact_loc_X"
    path = path_computer + f"results/mag_data"
    path_ex = path_computer + f"results/exact/mag_data"
    path_save = path_computer + f"figures/magnetization/"
    ylabel = "$\\sigma_{L/2}^x$"

elif args.what == "entropy":
    title = (
        f" ${args.where}-th$ Bond Entanglement Entropy: $L = {args.L}$ ;"
        + " $h_{ev}=$"
        + f" ${args.h_ev}$"
    )
    fname_what = f"{args.where}_bond_entropy"
    fname_ex_what = f"{args.where}_bond_exact_entropy"
    path = path_computer + f"results/entropy"
    path_ex = path_computer + f"results/exact/entropy"
    path_save = path_computer + f"figures/entropy/"
    ylabel = "entanglement von neumann entropy $(S_{\\chi})$"

elif args.what == "entropy_tot":
    title = (
        f"All Bonds Entanglement Entropy: $L = {args.L}$ ;"
        + " $h_{ev}=$"
        + f" ${args.h_ev}$"
    )
    fname_what = f"{args.where}_bond_entropy"
    path = path_computer + f"results/entropy"
    path_save = path_computer + f"figures/entropy/"
    xlabel = "bonds"
    x_ticks = range(args.L - 1)
    labels = range(1, args.L - 1 + 1)
    steps = len(x_ticks) // 5
    xticks = x_ticks[::steps]
    steps = len(labels) // 5
    xlabels = labels[::steps]
    y_ticks = np.arange(args.trotter_steps + 1)
    labels = delta * y_ticks
    steps = len(y_ticks) // 5
    yticks = y_ticks[::steps]
    steps = len(labels) // 5
    ylabels = labels[::steps]
    # x_grid = list(x_ticks).copy
    # y_grid = list(y_ticks).copy()
    # y_grid.pop(-1)
    # X,Y = np.meshgrid(np.asarray(x_grid), np.asarray(y_grid))
    X, Y = np.meshgrid(x_ticks, y_ticks)
    view_init = False

elif args.what == "mag_loc":
    title = (
        f"Local Magnetization Evolution: $L = {args.L}$ ;"
        + " $h_{ev}=$ "
        + f"${args.h_ev}$"
    )
    fname_what = f"mag_mps_loc"
    fname_ex_what = "mag_exact_loc"
    path = path_computer + f"results/mag_data"
    path_ex = path_computer + f"results/exact/mag_data"
    path_save = path_computer + f"figures/magnetization/"
    xlabel = "sites"
    x_ticks = range(args.L)
    labels = range(1, args.L + 1)
    steps = len(x_ticks) // 5
    xticks = x_ticks[::steps]
    steps = len(labels) // 5
    xlabels = labels[::steps]
    y_ticks = np.arange(args.trotter_steps + 1)
    labels = delta * y_ticks
    steps = len(y_ticks) // 5
    yticks = y_ticks[::steps]
    steps = len(labels) // 5
    ylabels = labels[::steps]
    X, Y = np.meshgrid(x_ticks, y_ticks)
    view_init = True

else:
    raise SyntaxError(
        "insert a valid result to plot: 'mag_tot', 'mag_z', 'mag_x', 'entropy', 'err_mag_tot', 'err_mag_z', 'err_mag_x', 'err_entropy'"
    )

if args.what in plot_val:
    fname_ex = f"{fname_ex_what}_{args.model}_L_{args.L}_midflip_{args.flip}_quench_{args.quench}_delta_{delta}_h_ev_{args.h_ev}"
    fname = f"{fname_what}_{args.model}_L_{args.L}_midflip_{args.flip}_quench_{args.quench}_delta_{delta}"
    second_part = f"_h_ev_{args.h_ev}"
    fname_save = f"{args.what}_{args.model}_L_{args.L}_midflip_{args.flip}_quench_{args.quench}_delta_{delta}"
    second_part = f"_h_ev_{args.h_ev}"
    plot_results_TEBD(
        title,
        for_array=args.chis,
        trotter_steps=args.trotter_steps,
        delta=delta,
        fname=fname,
        second_part=second_part,
        path=path,
        fname_save=fname_save,
        path_save=path_save,
        ylabel=ylabel,
        save=args.save,
        exact=args.exact,
        fname_ex=fname_ex,
        path_ex=path_ex,
        marker=args.marker,
        m_size=args.m_size,
        linewidth=args.linewidth,
        alpha=args.alpha,
        n_points=args.n_points,
        cmap=args.cmap,
    )

if args.what in plot_cmap:
    for chi in args.chis:
        title_fin = title + f" ; $\\chi = {chi}$"
        fname = f"{fname_what}_{args.model}_L_{args.L}_midflip_{args.flip}_quench_{args.quench}_delta_{delta}_chi_{chi}_h_ev_{args.h_ev}"
        fname_save = f"{args.what}_{args.model}_L_{args.L}_midflip_{args.flip}_quench_{args.quench}_delta_{delta}_chi_{chi}_h_ev_{args.h_ev}"
        plot_colormaps_evolution(
            title=title_fin,
            fname=fname,
            path=path,
            fname_save=fname_save,
            path_save=path_save,
            xlabel=xlabel,
            xticks=xticks,
            xlabels=xlabels,
            yticks=yticks,
            ylabels=ylabels,
            X=X,
            Y=Y,
            save=args.save,
            cmap=args.cmap,
            interpolation=args.interpolation,
            d=args.dim,
            view_init=view_init,
        )



import concurrent.futures
import os
from qs_mps.mps_class import MPS
from qs_mps.utils import tensor_shapes
import numpy as np

def ground_state_ANNNI_param(params):
    args_mps = params[0]
    h = params[1][0]
    k = params[1][1]
    chain = MPS(
        L=args_mps["L"],
        d=args_mps["d"],
        model=args_mps["model"],
        chi=args_mps["chi"],
        h=h,
        k=k,
        J=1
    )
    save = args_mps["save"]
    precision = args_mps["precision"]
    chain._random_state(seed=3, chi=args_mps["chi"], type_shape=args_mps["type_shape"])
    # tensor_shapes(chain.sites)
    chain.canonical_form(trunc_chi=True, trunc_tol=False)
    
    energy, entropy, schmidt_vals = chain.DMRG(
        trunc_tol=args_mps["trunc_tol"],
        trunc_chi=args_mps["trunc_chi"],
        where=args_mps["where"],
        bond=args_mps["bond"],
    )
    print(f"energy of h:{h:.{precision}f}, k:{k:.{precision}f} is:\n {energy}")
    print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")

    if save:
        chain.save_sites(args_mps["path"], args_mps["precision"])
    return energy, entropy, schmidt_vals


def ground_state_ANNNI_multpr(args_mps, multpr_param, cpu_percentage=90):
    max_workers = int(os.cpu_count() * (cpu_percentage / 100))
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        args = [[args_mps, [h,k]] for h in multpr_param[0] for k in multpr_param[1]]
        results = executor.map(ground_state_ANNNI_param, args)

    energies = []
    entropies = []
    precision = args_mps["precision"]
    i = 0
    j = 0
    for result in results:

        print(f"energy of h:{multpr_param[0][i]:.{precision}f}, k:{multpr_param[1][j]:.{precision}f} is:\n {result[0][-1]}")
        print(f"Schmidt values in the middle of the chain:\n {result[2]}")

        energies.append(result[0])
        entropies.append(result[1])
        if j == (len(multpr_param[1])-1):
            i += 1
            j = -1
        j += 1
        

    return energies, entropies


def ground_state_ANNNI(args_mps, multpr, param):
    if multpr:
        energies_param, entropies_param = ground_state_ANNNI_multpr(
            args_mps=args_mps, multpr_param=param
        )
    else:
        energies_param = []
        entropies_param = []
        schmidt_vals_param = []
        for h in param[0]:
            energies_h = []
            entropies_h = []
            schmidt_vals_h = []
            for k in param[1]:
                # if h == 0 and k > 0.48:
                params = [args_mps, [h,k]]
                energy, entropy, schmidt_vals = ground_state_ANNNI_param(params=params)
                energies_h.append(energy[-1])
                entropies_h.append(entropy)
                schmidt_vals_h.append(schmidt_vals)
            energies_param.append(energies_h)
            entropies_param.append(entropies_h)
            schmidt_vals_param.append(schmidt_vals_h)

    return energies_param, entropies_param, schmidt_vals_param



import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision, save_list_of_lists, access_txt
from qs_mps.applications.Z2.ground_state_multiprocessing import ground_state_Z2

# DENSITY MATRIX RENORMALIZATION GROUP to find ground states of the Z2 Pure Gauge Theory 
# changing the transverse field parameters in its dual formulation

parser = argparse.ArgumentParser(prog="gs_search_Z2")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Number of rungs per ladder", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument(
    "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="rectangular", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
)
parser.add_argument(
    "-mu", "--multpr", help="If True computes ground states with multiprocessing. By default False", action="store_true"
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=10,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-12,
    type=float,
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-v",
    "--save",
    help="Save the tensors. By default True",
    action="store_false",
)
parser.add_argument(
    "-tr",
    "--training",
    help="Save all the energies during the variational optimization. By default True",
    action="store_false",
)
parser.add_argument(
    "-i",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)

args = parser.parse_args()

# define the physical dimension
d = int(2**(args.l))

# define the interval of equally spaced values of external field
if args.interval == "lin":
    interval = np.linspace(args.h_i, args.h_f, args.npoints)
    num = (interval[-1] - interval[0]) / args.npoints
    precision = get_precision(num)
elif args.interval == "log":
    interval = np.logspace(args.h_i, args.h_f, args.npoints)
    precision = int(np.max([np.abs(args.h_f),np.abs(args.h_i)]))

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_tensor = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")


# ---------------------------------------------------------
# DMRG
# ---------------------------------------------------------
for L in args.Ls:
    # define the sector by looking of the given charges
    if args.charges_x == []:
        sector = "vacuum_sector"
        charges_x = None
        charges_y = None
    else:
        sector = f"{len(args.charges_x)}_particle(s)_sector"
        charges_x = args.charges_x
        charges_y = args.charges_y
    # where to look at for the entropy
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False

    # init_state = np.zeros((d))
    # init_state[0] = 1
    # init_state = init_state.reshape((1,d,1))
    # init_tensor = [init_state for _ in range(L-1)]
    init_tensor = []
    for chi in args.chis:  # L // 2 + 1
        args_mps = {
            "L": L,
            "d": d,
            "chi": chi,
            "type_shape": args.type_shape,
            "model": args.model,
            "trunc_tol": False,
            "trunc_chi": True,
            "where": args.where,
            "bond": args.bond,
            "path": path_tensor,
            "save": args.save,
            "precision": precision,
            "sector": sector,
            "charges_x": charges_x,
            "charges_y": charges_y,
            "n_sweeps": args.number_sweeps,
            "conv_tol": args.conv_tol,
            "training": args.training,
            "guess": init_tensor,
        }
        if __name__ == "__main__":
            energy_chi, entropy_chi, schmidt_vals_chi, t_chi = ground_state_Z2(
                args_mps=args_mps, multpr=args.multpr, param=interval
            )

            t_final = np.sum(t_chi)
            if t_final < 60:
                t_unit = "sec(s)"
            elif t_final > 60 and t_final < 3600:
                t_unit = "min(s)"
                t_final = t_final/60
            elif t_final > 3600:
                t_unit = "hour(s)"
                t_final = t_final/3600

            print(f"time of the whole search for chi={chi} is: {t_final} {t_unit}")
            if args.bond == False:
                args.where = "all"

            if args.training:
                energy_chi = np.asarray(energy_chi)
                energy_chi = energy_chi.reshape((len(interval),len(energy_chi[0])))
                print(energy_chi.shape)
                np.save(
                    f"{parent_path}/results/energy_data/energies_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    energy_chi,
                )
                for i in range(len(interval)):
                    np.save(f"{parent_path}/results/energy_data/energy_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}", energy_chi[i,-1])

            else:
                np.save(
                    f"{parent_path}/results/energy_data/energy_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    energy_chi,
                )

            save_list_of_lists(
                f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                entropy_chi,
            )
            # save_list_of_lists(
            #     f"{parent_path}/results/entropy_data/{args.where}_schmidt_vals_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
            #     schmidt_vals_chi,
            # )
            np.save(
                f"{parent_path}/results/entropy_data/{args.where}_schmidt_vals_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                schmidt_vals_chi,
            )
            if args.where == "all":
                entropy_mid = access_txt(
                    f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    (L-1) // 2,
                )
                np.savetxt(
                    f"{parent_path}/results/entropy_data/{args.L // 2}_bond_entropy_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    entropy_mid,
                )


from scipy.sparse import isspmatrix
import numpy as np

# from simsio import logger
from scipy.sparse.linalg import norm

__all__ = [
    "pause",
    "alert",
    "commutator",
    "anti_commutator",
    "check_commutator",
    "check_matrix",
    "check_hermitian",
]


def pause(phrase, debug):
    if not isinstance(phrase, str):
        raise TypeError(f"phrase should be a STRING, not a {type(phrase)}")
    if not isinstance(debug, bool):
        raise TypeError(f"debug should be a BOOL, not a {type(debug)}")
    if debug == True:
        # IT PROVIDES A PAUSE in a given point of the PYTHON CODE
        print("----------------------------------------------------")
        # Press the <ENTER> key to continue
        programPause = input(phrase)
        print("----------------------------------------------------")
        print("")


def alert(phrase, debug):
    if not isinstance(phrase, str):
        raise TypeError(f"phrase should be a STRING, not a {type(phrase)}")
    if not isinstance(debug, bool):
        raise TypeError(f"debug should be a BOOL, not a {type(debug)}")
    if debug == True:
        # IT PRINTS A PHRASE IN A GIVEN POINT OF A PYTHON CODE
        print("")
        print(phrase)


def commutator(A, B):
    # THIS FUNCTION COMPUTES THE COMMUTATOR of TWO SPARSE MATRICES
    if not isspmatrix(A):
        raise TypeError(f"A should be a csr_matrix, not a {type(A)}")
    if not isspmatrix(B):
        raise TypeError(f"B should be a csr_matrix, not a {type(B)}")
    return A * B - B * A


def anti_commutator(A, B):
    # THIS FUNCTION COMPUTES THE ANTI_COMMUTATOR of TWO SPARSE MATRICES
    if not isspmatrix(A):
        raise TypeError(f"A should be a csr_matrix, not a {type(A)}")
    if not isspmatrix(B):
        raise TypeError(f"B should be a csr_matrix, not a {type(B)}")
    return A * B + B * A


def check_commutator(A, B):
    # CHECKS THE COMMUTATION RELATIONS BETWEEN THE OPERATORS A AND B
    if not isspmatrix(A):
        raise TypeError(f"A should be a csr_matrix, not a {type(A)}")
    if not isspmatrix(B):
        raise TypeError(f"B should be a csr_matrix, not a {type(B)}")
    norma = norm(A * B - B * A)
    norma_max = max(norm(A * B + B * A), norm(A), norm(B))
    ratio = norma / norma_max
    # check=(AB!=BA).nnz
    if ratio > 10 ** (-15):
        print("    ERROR: A and B do NOT COMMUTE")
        print("    NORM", norma)
        print("    RATIO", ratio)
    print("")


def check_matrix(A, B):
    # CHEKS THE DIFFERENCE BETWEEN TWO SPARSE MATRICES
    if not isspmatrix(A):
        raise TypeError(f"A should be a csr_matrix, not a {type(A)}")
    if not isspmatrix(B):
        raise TypeError(f"B should be a csr_matrix, not a {type(B)}")
    if A.shape != B.shape:
        raise ValueError(f"Shape mismatch between : A {A.shape} & B: {B.shape}")
    norma = norm(A - B)
    norma_max = max(norm(A + B), norm(A), norm(B))
    ratio = norma / norma_max
    if ratio > 1e-5:
        print("    ERROR: A and B are DIFFERENT MATRICES")
        raise ValueError(f"    NORM {norma}, RATIO {ratio}")
    return ratio


def check_hermitian(A):
    if not isspmatrix(A):
        raise TypeError(f"A should be a csr_matrix, not a {type(A)}")
    # Get the Hermitian
    print("CHECK HERMITICITY")
    A_dag = A.getH()
    check_matrix(A, A_dag)
